#ifndef DENSE_H
#define DENSE_H

#include "kktdirect.h"

// Selected dense kernels from BLAS and LAPACK for supernodal mode
// Make sure you #define one of USE_FORTRAN_BLAS, USE_CBLAS, or USE_NAIVE_BLAS,
// and that you #define one of USE_FORTRAN_LAPACK, USE_CLAPACK, or USE_NAIVE_LAPACK.

#ifdef USE_FORTRAN_BLAS
//============================================ FORTRAN BLAS ==================================================
// I assume lowercase with an underscore suffix convention; if your platform does things differently you will
// need to modify this!

extern "C" {
void dgemm_(const char *TRANSA, const char *TRANSB, const KKT_INT *M, const KKT_INT *N, const KKT_INT *K,
            const double *ALPHA, const double *A, const KKT_INT *LDA, const double *B, const KKT_INT *LDB,
            const double *BETA, double *C, const KKT_INT *LDC);
}

namespace BLAS{
   inline void dgemm_NN(KKT_INT m, KKT_INT n, KKT_INT k, const double *A, const double *B, double *C)
   {
      const double one=1, zero=0;
      dgemm_("N", "N", &m, &n, &k, &one, A, &m, B, &k, &zero, C, &m);
   }

   inline void dgemm_TN(KKT_INT m, KKT_INT n, KKT_INT k, const double *A, const double *B, double *C)
   {
      const double one=1, zero=0;
      dgemm_("T", "N", &m, &n, &k, &one, A, &k, B, &k, &zero, C, &m);
   }

   inline void dgemm_subNN(KKT_INT m, KKT_INT n, KKT_INT k, const double *A, const double *B, double *C)
   {
      const double one=1, minus_one=-1;
      dgemm_("T", "N", &m, &n, &k, &minus_one, A, &m, B, &k, &one, C, &m);
   }
};

#elif defined USE_CBLAS
//=============================================== cblas ======================================================
#include "cblas.h"

namespace BLAS{
   inline void dgemm_NN(KKT_INT m, KKT_INT n, KKT_INT k, const double *A, const double *B, double *C)
   {
      cblas_dgemm(CblasColMajor, CblasNoTrans, CblasNoTrans, m, n, k, 1.0, A, m, B, k, 0.0, C, m);
   }

   inline void dgemm_TN(KKT_INT m, KKT_INT n, KKT_INT k, const double *A, const double *B, double *C)
   {
      cblas_dgemm(CblasColMajor, CblasTrans, CblasNoTrans, m, n, k, 1.0, A, k, B, k, 0.0, C, m);
   }

   inline void dgemm_subNN(KKT_INT m, KKT_INT n, KKT_INT k, const double *A, const double *B, double *C)
   {
      cblas_dgemm(CblasColMajor, CblasNoTrans, CblasNoTrans, m, n, k, -1.0, A, m, B, k, 1.0, C, m);
   }
};

#elif defined USE_NAIVE_BLAS
//============================================== Naive BLAS ==================================================
// NOTE: this is really just a fallback option if you really cannot get your hands on optimized BLAS. Do not
// expect good performance from these naive routines!

#include <cstring>

namespace BLAS{
   void dgemm_NN(KKT_INT m, KKT_INT n, KKT_INT k, const double * A, const double * B, double * C)
   {
      std::memset(C, 0, m*n*sizeof(double));
      // C[i+m*j]=sum_p A[i+m*p]*B[p+k*j]
      KKT_INT i, j, p;
      for(j=0; j<n; ++j)
         for(p=0; p<k; ++p){
            double Bpj=B[p+k*j];
            for(i=0; i<m; ++i)
               C[i+m*j]+=A[i+m*p]*Bpj;
         }
   }

   void dgemm_TN(KKT_INT m, KKT_INT n, KKT_INT k, const double * A, const double * B, double * C)
   {
      // C[i+m*j]=sum_p A[p+k*i]*B[p+k*j]
      KKT_INT i, j, p;
      for(j=0; j<n; ++j)
         for(i=0; i<m; ++i){
            double Cij=0;
            for(p=0; p<k; ++p)
               Cij+=A[p+k*i]*B[p+k*j];
            C[i+m*j]=Cij;
         }
   }

   void dgemm_subNN(KKT_INT m, KKT_INT n, KKT_INT k, const double * A, const double * B, double * C)
   {
      // C[i+m*j]-=sum_p A[i+m*p]*B[p+k*j]
      KKT_INT i, j, p;
      for(j=0; j<n; ++j)
         for(p=0; p<k; ++p){
            double Bpj=B[p+k*j];
            for(i=0; i<m; ++i)
               C[i+m*j]-=A[i+m*p]*Bpj;
         }
   }
};
#endif

#ifdef USE_FORTRAN_LAPACK
//========================================== FORTRAN LAPACK ================================================
// Note: should be compatible with f2c'd CLAPACK as well (just don't bother including the clapack.h header file)
// and Apple's vecLib LAPACK.

extern "C" {
void dpotrf_(const char *UPLO, KKT_INT *N, double *A, KKT_INT *LDA, KKT_INT *INFO);
void dpotrs_(const char *UPLO, KKT_INT *N, KKT_INT *NRHS, double *A, KKT_INT *LDA,
             double *B, KKT_INT *LDB, KKT_INT *INFO);
void dsytrf_(const char *UPLO, KKT_INT *N, double *A, KKT_INT *LDA, KKT_INT *IPIV,
             double *WORK, KKT_INT *LWORK, KKT_INT *INFO);
void dsytrs_(const char *UPLO, KKT_INT *N, KKT_INT *NRHS, double *A, KKT_INT *LDA, KKT_INT *IPIV,
             double *B, KKT_INT *LDB, KKT_INT *INFO);
}

namespace LAPACK{
   inline void factor_symmetric_positive_definite(KKT_INT n, double *A, KKT_INT &info)
   {
      dpotrf_("L", &n, A, &n, &info);
   }

   inline void solve_symmetric_positive_definite_with_factorization(KKT_INT n, KKT_INT nrhs,
                                                                    const double *factorization,
                                                                    double *righthandside, KKT_INT &info)
   {
      dpotrs_("L", &n, &nrhs, (double*)factorization, &n, righthandside, &n, &info);
   }

   inline void factor_symmetric_indefinite(KKT_INT n, double *A, KKT_INT *ipiv, double *work, KKT_INT worksize, KKT_INT &info)
   {
      dsytrf_("L", &n, A, &n, ipiv, work, &worksize, &info);
   }
   
   inline void solve_symmetric_indefinite_with_factorization(KKT_INT n, KKT_INT nrhs, const double *factorization, const KKT_INT *ipiv,
                                                             double *righthandside, KKT_INT &info)
   {
      dsytrs_("L", &n, &nrhs, (double*)factorization, &n, (KKT_INT*)ipiv, righthandside, &n, &info);
   }
};

#elif defined USE_NAIVE_LAPACK
//============================================ Naive LAPACK ================================================
// NOTE: this is really just a fallback option if you really cannot get your hands on proper LAPACK.

#include <string.h>

namespace LAPACK{

   inline void factor_symmetric_positive_definite(KKT_INT n, double *A, KKT_INT &info)
   {
      info=0;
      KKT_INT i, j, k;
      double d, over_d;
      for(i=0; i<n; ++i){
          d=A[i+n*i];
          for(j=0; j<i; ++j)
              d-=A[i+n*j]*A[i+n*j]*A[j+n*j];
          if(d<=0){
              A[i+n*i]=1; // default pivot - could be arbitrarily bad of course!
              for(j=i+1; j<n; ++j) A[j+n*i]=0; // zero out this column of L
              if(info==0) info=i+1; // provide info only on first bad pivot
          }else{
              A[i+n*i]=d;
              for(k=0; k<i; ++k){
                  double s=A[k+n*k]*A[i+n*k];
                  for(j=i+1; j<n; ++j)
                      A[j+n*i]-=A[j+n*k]*s;
              }
              over_d=1/d;
              for(j=i+1; j<n; ++j) A[j+n*i]*=over_d;
          }
      }
   }

   inline void solve_symmetric_positive_definite_with_factorization(KKT_INT n, KKT_INT nrhs,
                                                                    const double *L,
                                                                    double *x, KKT_INT &info)
   {
       info=0;
       KKT_INT i, j, k;
       // solve in-place with L and divide by D as we go
       for(i=0; i<n; ++i){
           for(j=0; j<nrhs; ++j){
               double xij=x[i+n*j];
               for(k=i+1; k<n; ++k)
                   x[k+n*j]-=L[k+n*i]*xij;
               x[i+n*j]=xij/L[i+n*i];
           }
       }
       // solve in-place with L^T
       for(i=n-1; i>=0; --i){
           for(j=0; j<nrhs; ++j){
               double xij=x[i+n*j];
               for(k=i+1; k<n; ++k)
                   xij-=L[k+n*i]*x[k+n*j];
               x[i+n*j]=xij;
           }
       }
   }

   // for now assume no pivoting required
   inline void factor_symmetric_indefinite(KKT_INT n, double *A, KKT_INT *ipiv, double *work, KKT_INT worksize, KKT_INT &info)
   {
      info=0;
      KKT_INT i, j, k;
      double d, over_d;
      for(i=0; i<n; ++i){
         d=A[i+n*i];
         for(j=0; j<i; ++j)
            d-=A[i+n*j]*A[i+n*j]*A[j+n*j];
         if(d==0){
            A[i+n*i]=1; // default pivot - could be arbitrarily bad of course!
            for(j=i+1; j<n; ++j) A[j+n*i]=0; // zero out this column of L
            if(info==0) info=i+1; // only give info on first bad pivot
         }else{
            A[i+n*i]=d;
            for(k=0; k<i; ++k){
               double s=A[k+n*k]*A[i+n*k];
               for(j=i+1; j<n; ++j)
                  A[j+n*i]-=A[j+n*k]*s;
            }
            over_d=1/d;
            for(j=i+1; j<n; ++j) A[j+n*i]*=over_d;
         }
      }
   }
  
   // since we assumed no pivoting above, we can use the same solve routine
   inline void solve_symmetric_indefinite_with_factorization(KKT_INT n, KKT_INT nrhs, const double *L, const KKT_INT *ipiv,
                                                             double *x, KKT_INT &info)
   {
      solve_symmetric_positive_definite_with_factorization(n, nrhs, L, x, info);
   }
};
#endif

#endif
