#include <assert.h>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <vector>
#include "kktdirect.h"

struct Problem
{
   KKT_INT n;
   KKT_INT *colstart;
   KKT_INT *rowindex;
   double *value;
   char *constrained;
   double *exactsolution;
   double solution_norm;
   double *righthandside;
   double rhs_norm;
};

struct Ordering
{
   KKT_INT *original;
   KKT_INT *post;
   KKT_INT *inverse;
   KKT_INT *etreeparent;
};

double mymax(double a, double b)
{
   return (a>b) ? a : b;
}

void print_matrix(KKT_INT n, KKT_INT *colstart, KKT_INT *rowindex, double *value, KKT_INT *ordering=0)
{
   std::vector<double> dense(n*n, 0);
   for(KKT_INT j=0; j<n; ++j){
      for(KKT_INT i=colstart[j]; i<colstart[j+1]; ++i){
         int row, column;
         if(!ordering){
            row=rowindex[i];
            column=j;
         }else{
            row=ordering[rowindex[i]];
            column=ordering[j];
         }
         dense[row+column*n]=value[i];
      }
   }
   for(KKT_INT j=0; j<n; ++j){
      for(KKT_INT i=0; i<n; ++i){
         std::printf(" % 7.4f", dense[j+i*n]);
      }
      std::printf("\n");
   }
}

void print_supernodal_factor(KKT_INT n, KKT_INT numsupernodes, KKT_INT *supernodestart,
                             KKT_INT *supercolstart, KKT_INT *superrowindex, KKT_INT *factorcolstart,
                             double *factorvalue)
{
   std::vector<double> dense(n*n, 0);
   for(KKT_INT i=0; i<n; ++i){
      dense[i*(n+1)]=1;
   }
   for(KKT_INT s=0; s<numsupernodes; ++s){
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      for(KKT_INT j=supercolstart[s]; j<supercolstart[s+1]; ++j){
         int row=superrowindex[j];
         for(KKT_INT i=0; i<supersize; ++i){
            int column=supernodestart[s]+i;
            dense[row+column*n]=factorvalue[factorcolstart[s]+(j-supercolstart[s])+i];
         }
      }
   }
   for(KKT_INT j=0; j<n; ++j){
      for(KKT_INT i=0; i<n; ++i){
         std::printf(" % 7.4f", dense[j+i*n]);
      }
      std::printf("\n");
   }
}

void print_supernodal_diag(KKT_INT n, KKT_INT numsupernodes, KKT_INT *supernodestart,
                           KKT_INT *diagblockstart, double *diagvalue)
{
   std::vector<double> dense(n*n, 0);
   for(KKT_INT s=0; s<numsupernodes; ++s){
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      for(KKT_INT j=0; j<supersize; ++j) for(KKT_INT i=j; i<supersize; ++i){
         dense[(supernodestart[s]+i)+(supernodestart[s]+j)*n]=diagvalue[diagblockstart[s]+i+j*supersize];
      }
   }
   for(KKT_INT j=0; j<n; ++j){
      for(KKT_INT i=0; i<n; ++i){
         std::printf(" % 7.4f", dense[j+i*n]);
      }
      std::printf("\n");
   }
}

void initialize_problem(Problem &problem)
{
   std::printf("READING *****************************************\n");

   std::scanf("%d", &problem.n);
   problem.colstart=(KKT_INT*)std::calloc(problem.n+1, sizeof(KKT_INT));
   for(KKT_INT i=0; i<problem.n+1; ++i){
      std::scanf("%d", &problem.colstart[i]);
   }
   problem.rowindex=(KKT_INT*)std::calloc(problem.colstart[problem.n], sizeof(KKT_INT));
   for(KKT_INT i=0; i<problem.colstart[problem.n]; ++i){
      std::scanf("%d", &problem.rowindex[i]);
   }
   problem.value=(double*)std::calloc(problem.colstart[problem.n], sizeof(double));
   for(KKT_INT i=0; i<problem.colstart[problem.n]; ++i){
      std::scanf("%lf", &problem.value[i]);
   }

   problem.constrained=(char*)std::calloc(problem.n, sizeof(char));
   KKT_detect_constraints(problem.n, problem.colstart, problem.rowindex, problem.value,
                          problem.constrained);

   KKT_INT num_constraints=0;
   for(KKT_INT i=0; i<problem.n; ++i){
      if(problem.constrained[i]) ++num_constraints;
   }
   std::printf("Matrix: dimension %d, %d nonzeros, %d constraints\n", problem.n, problem.colstart[problem.n], num_constraints);

   problem.exactsolution=(double*)std::calloc(problem.n, sizeof(double));
   problem.righthandside=(double*)std::calloc(problem.n, sizeof(double));
   problem.solution_norm=0;
   for(KKT_INT i=0; i<problem.n; ++i){
      problem.exactsolution[i]=rand()/(double)RAND_MAX-0.5;
      problem.solution_norm=mymax(problem.solution_norm, fabs(problem.exactsolution[i]));
   }
   problem.rhs_norm=0;
   for(KKT_INT i=0; i<problem.n; ++i){ // using symmetry of matrix to do matrix-vector product...
      double bi=0;
      for(KKT_INT j=problem.colstart[i]; j<problem.colstart[i+1]; ++j){
         bi+=problem.value[j]*problem.exactsolution[problem.rowindex[j]];
      }
      problem.righthandside[i]=bi;
      problem.rhs_norm=mymax(problem.rhs_norm, fabs(problem.righthandside[i]));
   }
}

void run_ordering(const Problem &problem, Ordering &ordering)
{
   std::printf("ORDERING ****************************************\n");

   ordering.original=(KKT_INT*)std::calloc(problem.n, sizeof(KKT_INT));
   ordering.post=(KKT_INT*)std::calloc(problem.n, sizeof(KKT_INT));
   ordering.inverse=(KKT_INT*)std::calloc(problem.n, sizeof(KKT_INT));
   ordering.etreeparent=(KKT_INT*)std::calloc(problem.n, sizeof(KKT_INT));

   KKT_minimum_degree(problem.n, problem.colstart, problem.rowindex, problem.constrained, 1,
                      ordering.original);
   if(!KKT_check_constraints_satisified_by_ordering(problem.n, problem.colstart, problem.rowindex, problem.constrained, ordering.original)){
      std::printf("error in minimum degree: doesn't satisfy constraints\n");
   }

   KKT_postorder(problem.n, problem.colstart, problem.rowindex, ordering.original,
                 ordering.post, ordering.inverse, ordering.etreeparent);
   if(!KKT_check_constraints_satisified_by_ordering(problem.n, problem.colstart, problem.rowindex, problem.constrained, ordering.post)){
      std::printf("error in post ordering: doesn't satisfy constraints\n");
   }
}

void run_uplooking(const Problem &problem, const Ordering &ordering)
{
   std::printf("UPLOOKING TEST **********************************\n");

   KKT_INT *factorcolstart=(KKT_INT*)std::calloc(problem.n+1, sizeof(KKT_INT));

   KKT_symbolically_factor_uplooking(problem.n, problem.colstart, problem.rowindex,
                                     ordering.post, ordering.inverse, ordering.etreeparent,
                                     factorcolstart);
   std::printf("%d + %d = %d nonzeros in factorization\n", factorcolstart[problem.n], problem.n, factorcolstart[problem.n]+problem.n);

   KKT_INT *factorrowindex=(KKT_INT*)std::calloc(factorcolstart[problem.n], sizeof(KKT_INT));
   double *factorvalue=(double*)std::calloc(factorcolstart[problem.n], sizeof(double));
   double *diagvalue=(double*)std::calloc(problem.n, sizeof(double));

   int result=KKT_numerically_factor_uplooking(problem.n, problem.colstart, problem.rowindex, problem.value,
                                               ordering.post, ordering.inverse, ordering.etreeparent, factorcolstart,
                                               factorrowindex, factorvalue, diagvalue);
   std::printf("result of numerical factorization: %d\n", result);

   double *solution=(double*)std::calloc(problem.n, sizeof(double));
   double *matrix_times_solution=(double*)std::calloc(problem.n, sizeof(double));

   KKT_solve_with_pointwise_factorization(problem.n, ordering.post, factorcolstart, factorrowindex, factorvalue, diagvalue, problem.righthandside,
                                          solution);

   for(KKT_INT i=0; i<problem.n; ++i){
      double bi=0;
      for(KKT_INT j=problem.colstart[i]; j<problem.colstart[i+1]; ++j){
         bi+=problem.value[j]*solution[problem.rowindex[j]];
      }
      matrix_times_solution[i]=bi;
   }
   double maxerror=0, maxresidual=0;
   for(KKT_INT i=0; i<problem.n; ++i){
      maxerror=mymax(maxerror, std::fabs(solution[i]-problem.exactsolution[i]));
      maxresidual=mymax(maxresidual, std::fabs(problem.righthandside[i]-matrix_times_solution[i]));
   }
   std::printf("relative error=%lg\n", maxerror/problem.solution_norm);
   std::printf("relative residual=%lg\n", maxresidual/problem.rhs_norm);

   std::free(factorcolstart);
   std::free(factorrowindex);
   std::free(factorvalue);
   std::free(diagvalue);
   std::free(solution);
   std::free(matrix_times_solution);
}

void run_signed_cholesky(const Problem &problem, const Ordering &ordering)
{
   std::printf("SIGNED CHOLESKY TEST ****************************\n");

   KKT_INT *factorcolstart=(KKT_INT*)std::calloc(problem.n+1, sizeof(KKT_INT));

   KKT_symbolically_factor_signed_cholesky(problem.n, problem.colstart, problem.rowindex,
                                     ordering.post, ordering.inverse, ordering.etreeparent,
                                     factorcolstart);
   std::printf("%d nonzeros in factorization\n", factorcolstart[problem.n]);

   KKT_INT *factorrowindex=(KKT_INT*)std::calloc(factorcolstart[problem.n], sizeof(KKT_INT));
   double *factorvalue=(double*)std::calloc(factorcolstart[problem.n], sizeof(double));

   int result=KKT_numerically_factor_signed_cholesky(problem.n, problem.colstart, problem.rowindex, problem.value, problem.constrained,
                                                     ordering.post, ordering.inverse, ordering.etreeparent, factorcolstart,
                                                     factorrowindex, factorvalue);
   std::printf("result of numerical factorization: %d\n", result);

   double *solution=(double*)std::calloc(problem.n, sizeof(double));
   double *matrix_times_solution=(double*)std::calloc(problem.n, sizeof(double));

   KKT_solve_with_signed_cholesky(problem.n, problem.constrained, ordering.post, factorcolstart, factorrowindex, factorvalue, problem.righthandside,
                                  solution);

   for(KKT_INT i=0; i<problem.n; ++i){
      double bi=0;
      for(KKT_INT j=problem.colstart[i]; j<problem.colstart[i+1]; ++j){
         bi+=problem.value[j]*solution[problem.rowindex[j]];
      }
      matrix_times_solution[i]=bi;
   }
   double maxerror=0, maxresidual=0;
   for(KKT_INT i=0; i<problem.n; ++i){
      maxerror=mymax(maxerror, std::fabs(solution[i]-problem.exactsolution[i]));
      maxresidual=mymax(maxresidual, std::fabs(problem.righthandside[i]-matrix_times_solution[i]));
   }
   std::printf("relative error=%lg\n", maxerror/problem.solution_norm);
   std::printf("relative residual=%lg\n", maxresidual/problem.rhs_norm);

   std::free(factorcolstart);
   std::free(factorrowindex);
   std::free(factorvalue);
   std::free(solution);
   std::free(matrix_times_solution);
}

void run_supernodal(const Problem &problem, const Ordering &ordering)
{
   std::printf("SUPERNODAL TEST *********************************\n");

   KKT_INT numsupernodes=0;
   KKT_INT *supernodestart=(KKT_INT*)std::calloc(problem.n+1, sizeof(KKT_INT));
   KKT_INT *supernodemap=(KKT_INT*)std::calloc(problem.n, sizeof(KKT_INT));
   KKT_INT *supercolstart=(KKT_INT*)std::calloc(problem.n+1, sizeof(KKT_INT));
   KKT_INT *factorcolstart=(KKT_INT*)std::calloc(problem.n+1, sizeof(KKT_INT));
   KKT_INT *diagblockstart=(KKT_INT*)std::calloc(problem.n+1, sizeof(KKT_INT));

   KKT_INT max_supernode_size=64;
   KKT_symbolically_factor_supernodal_initial(problem.n, problem.colstart, problem.rowindex,
                                              ordering.post, ordering.inverse, ordering.etreeparent, max_supernode_size,
                                              &numsupernodes, supernodestart, supernodemap,
                                              supercolstart, factorcolstart, diagblockstart);
   std::printf("%d supernodes (average size %.2f)\n", numsupernodes, problem.n/(double)numsupernodes);
   std::printf("%d + %d = %d nonzeros in factorization\n", factorcolstart[numsupernodes], diagblockstart[numsupernodes], factorcolstart[numsupernodes]+diagblockstart[numsupernodes]);

   KKT_INT *superrowindex=(KKT_INT*)std::calloc(supercolstart[numsupernodes], sizeof(KKT_INT));

   KKT_symbolically_factor_supernodal_final(problem.n, problem.colstart, problem.rowindex,
                                            ordering.post, ordering.inverse, ordering.etreeparent,
                                            numsupernodes, supernodestart, supernodemap, supercolstart, 
                                            superrowindex);
   std::printf("finished second stage of symbolic factorization\n");

   double *factorvalue=(double*)std::calloc(factorcolstart[numsupernodes], sizeof(double));
   double *diagvalue=(double*)std::calloc(diagblockstart[numsupernodes], sizeof(double));
   KKT_INT *diagpiv=(KKT_INT*)std::calloc(problem.n, sizeof(KKT_INT));

   int result=KKT_numerically_factor_leftlooking_supernodal(problem.n, problem.colstart, problem.rowindex, problem.value,
                                                            ordering.post, ordering.inverse, ordering.etreeparent,
                                                            numsupernodes, supernodestart, supernodemap,
                                                            supercolstart, superrowindex, factorcolstart, factorvalue,
                                                            diagblockstart, diagvalue, diagpiv);
   std::printf("result of numerical factorization: %d\n", result);

   double *solution=(double*)std::calloc(problem.n, sizeof(double));
   double *matrix_times_solution=(double*)std::calloc(problem.n, sizeof(double));

   KKT_solve_with_supernodal_factorization(problem.n, ordering.post, numsupernodes, supernodestart, supernodemap,
                                           supercolstart, superrowindex, factorcolstart, factorvalue,
                                           diagblockstart, diagvalue, diagpiv,
                                           1, problem.righthandside,
                                           solution);

   for(KKT_INT i=0; i<problem.n; ++i){
      double bi=0;
      for(KKT_INT j=problem.colstart[i]; j<problem.colstart[i+1]; ++j){
         bi+=problem.value[j]*solution[problem.rowindex[j]];
      }
      matrix_times_solution[i]=bi;
   }
   double maxerror=0, maxresidual=0;
   for(KKT_INT i=0; i<problem.n; ++i){
      maxerror=mymax(maxerror, std::fabs(solution[i]-problem.exactsolution[i]));
      maxresidual=mymax(maxresidual, std::fabs(problem.righthandside[i]-matrix_times_solution[i]));
   }
   std::printf("relative error=%lg\n", maxerror/problem.solution_norm);
   std::printf("relative residual=%lg\n", maxresidual/problem.rhs_norm);

   std::free(supernodestart);
   std::free(supernodemap);
   std::free(supercolstart);
   std::free(factorcolstart);
   std::free(diagblockstart);
   std::free(superrowindex);
   std::free(factorvalue);
   std::free(diagvalue);
   std::free(diagpiv);
   std::free(solution);
   std::free(matrix_times_solution);
}

void run_supernodal_signed_cholesky(const Problem &problem, Ordering &ordering)
{
   std::printf("SUPERNODAL SIGNED CHOLESKY TEST *****************\n");

   KKT_INT numsupernodes=0;
   KKT_INT *supernodestart=(KKT_INT*)std::calloc(problem.n+1, sizeof(KKT_INT));
   KKT_INT *supernodemap=(KKT_INT*)std::calloc(problem.n, sizeof(KKT_INT));
   KKT_INT *supercolstart=(KKT_INT*)std::calloc(problem.n+1, sizeof(KKT_INT));
   KKT_INT *factorcolstart=(KKT_INT*)std::calloc(problem.n+1, sizeof(KKT_INT));
   KKT_INT *diagblockstart=(KKT_INT*)std::calloc(problem.n+1, sizeof(KKT_INT));

   KKT_INT max_supernode_size=64;
   KKT_symbolically_factor_supernodal_initial_for_signed_cholesky(problem.n, problem.colstart, problem.rowindex,
                                                                  problem.constrained, max_supernode_size,
                                                                  ordering.post, ordering.inverse, ordering.etreeparent,
                                                                  &numsupernodes, supernodestart, supernodemap,
                                                                  supercolstart, factorcolstart, diagblockstart);
   std::printf("%d supernodes (average size %.2f)\n", numsupernodes, problem.n/(double)numsupernodes);
   std::printf("%d + %d = %d nonzeros in factorization\n", factorcolstart[numsupernodes], diagblockstart[numsupernodes], factorcolstart[numsupernodes]+diagblockstart[numsupernodes]);

   if(!KKT_check_constraints_satisified_by_ordering(problem.n, problem.colstart, problem.rowindex, problem.constrained, ordering.post)){
      std::printf("error in modified ordering: doesn't satisfy constraints\n");
   }

   KKT_INT *superrowindex=(KKT_INT*)std::calloc(supercolstart[numsupernodes], sizeof(KKT_INT));

   KKT_symbolically_factor_supernodal_final(problem.n, problem.colstart, problem.rowindex,
                                            ordering.post, ordering.inverse, ordering.etreeparent,
                                            numsupernodes, supernodestart, supernodemap, supercolstart, 
                                            superrowindex);
   std::printf("finished second stage of symbolic factorization\n");

   double *factorvalue=(double*)std::calloc(factorcolstart[numsupernodes], sizeof(double));
   double *diagvalue=(double*)std::calloc(diagblockstart[numsupernodes], sizeof(double));

   int result=KKT_numerically_factor_supernodal_signed_cholesky(problem.n, problem.colstart, problem.rowindex, problem.value,
                                                                ordering.post, ordering.inverse, ordering.etreeparent,
                                                                problem.constrained,
                                                                numsupernodes, supernodestart, supernodemap,
                                                                supercolstart, superrowindex, factorcolstart, factorvalue,
                                                                diagblockstart, diagvalue);
   std::printf("result of numerical factorization: %d\n", result);

   double *solution=(double*)std::calloc(problem.n, sizeof(double));
   double *matrix_times_solution=(double*)std::calloc(problem.n, sizeof(double));

   KKT_solve_with_supernodal_signed_cholesky(problem.n, problem.constrained, ordering.post,
                                             numsupernodes, supernodestart, supernodemap,
                                             supercolstart, superrowindex, factorcolstart, factorvalue,
                                             diagblockstart, diagvalue,
                                             1, problem.righthandside,
                                             solution);

   for(KKT_INT i=0; i<problem.n; ++i){
      double bi=0;
      for(KKT_INT j=problem.colstart[i]; j<problem.colstart[i+1]; ++j){
         bi+=problem.value[j]*solution[problem.rowindex[j]];
      }
      matrix_times_solution[i]=bi;
   }
   double maxerror=0, maxresidual=0;
   for(KKT_INT i=0; i<problem.n; ++i){
      maxerror=mymax(maxerror, std::fabs(solution[i]-problem.exactsolution[i]));
      maxresidual=mymax(maxresidual, std::fabs(problem.righthandside[i]-matrix_times_solution[i]));
   }
   std::printf("relative error=%lg\n", maxerror/problem.solution_norm);
   std::printf("relative residual=%lg\n", maxresidual/problem.rhs_norm);

   std::free(supernodestart);
   std::free(supernodemap);
   std::free(supercolstart);
   std::free(factorcolstart);
   std::free(diagblockstart);
   std::free(superrowindex);
   std::free(factorvalue);
   std::free(diagvalue);
   std::free(solution);
   std::free(matrix_times_solution);
}

int main()
{
   Problem problem;
   Ordering ordering;

   initialize_problem(problem);
   run_ordering(problem, ordering);

   run_uplooking(problem, ordering);
   run_signed_cholesky(problem, ordering);
   run_supernodal(problem, ordering);
   run_supernodal_signed_cholesky(problem, ordering);

   return 0;
}

