#include "kktdirect.h"
#include "kktsupport.h"

//=================================== KKTOrdering ============================
// The guts of the minimum cost ordering.
// This is in fact just a virtual base class; you need to subclass
// this to define the cost metric for a particular minimum-degree-like algorithm.

struct KKTOrdering
{
   // initializes quotient graph, and pre-orders isolated nodes
   KKTOrdering(KKT_INT n_, const KKT_INT *colstart, const KKT_INT *rowindex, const char *constrained_, int safely_handle_dense_nodes, KKT_INT *ordering_);
   virtual ~KKTOrdering(void) {}

   void do_the_ordering(void);

   protected:
   void initialize_ordering(void);
   void eliminate_node(KKT_INT i);
   void check_constrained_node(KKT_INT j);
   void find_supernodes_around_eliminated_node(KKT_INT i);
   virtual void initialize_cost(KKT_INT i) = 0;
   virtual void update_cost(KKT_INT j) = 0;
   virtual void cheap_update_cost(KKT_INT eliminated_node, KKT_UINT previous_size) = 0;
   KKT_INT find_super_root(KKT_INT &j); // note: modifies j to point to the root (and returns the root)
   void print_state(void);

   // state of the ordering

   KKT_INT n; // number of nodes
   KKT_INT order_index; // how many nodes we have ordered/eliminated so far
   KKT_INT *ordering; // the vector we're recording the ordering in
   std::vector<std::vector<KKT_INT> > adj; // for each node, current adjacency list in quotient graph
   std::vector<KKT_UINT> cost; // for each node, its cost
   std::vector<bool> accurate_cost; // set to false if the cost hasn't been updated
   PriorityQueue pq; // for selecting the node of minimum cost: contains all nodes that can currently be ordered
   const char *constrained; // for each node, if it's a constrained KKT variable or not
   std::vector<bool> eliminated; // for each node, indicates if it has already been eliminated
   std::vector<KKT_INT> superparent; // for each node, a pointer to follow up to find the root of its supernode
                                 // (-1 for a root itself). If the node has been eliminated, its "supernode" swells
                                 // to include all merged eliminated nodes in a quotient node.
   std::vector<std::vector<KKT_INT> > superlist; // for each node, if its an uneliminated supernode root, the list of
                                             // all nodes it represents (including itself).
                                             // One important invariant: the constrained status of all the nodes in
                                             // a supernode is enforced to be identical.
   Marker marker; // used for a variety of purposes
   std::vector<KKT_INT> unconstrained_at_end, constrained_at_end; // remove dense rows/columns from the get-go and put them last

   std::vector<KKT_INT> number_of_cost_updates;
};

KKTOrdering::
KKTOrdering(KKT_INT n_, const KKT_INT *colstart, const KKT_INT *rowindex, const char *constrained_, int safely_handle_dense_nodes, KKT_INT *ordering_)
   : n(n_),
     order_index(0), ordering(ordering_),
     adj(n), cost(n,0), accurate_cost(n,false), pq(cost),
     constrained(constrained_),
     eliminated(n, false), // nothing eliminated at the start
     superparent(n, -1), // the value -1 indicates each node begins as its own supernode root
     superlist(n), // will be filled in below
     marker(n),
     number_of_cost_updates(n, 0)
{
   assert(constrained);
   // The goal here is to translate the input sparse matrix structure (colstart, rowindex) into a graph.
   // However, we do a little extra processing (eliminating isolated nodes and stripping out overly dense nodes)
   // before conversion.

   // Eliminate any isolated nodes instead of bothering to put them in the graph,
   // and remove dense rows/columns to be ordered at the end.
   //KKT_INT dense_threshhold=n+1;
   //KKT_INT dense_threshhold=10;
   KKT_INT dense_threshhold=std::min(n-2, std::max(16, (KKT_INT)(10*std::sqrt((double)n))));
   for(KKT_INT i=0; i<n; ++i){
      if(eliminated[i]) continue;
      // Check if this is an isolated node that can be ordered right away.
      if(colstart[i+1]==colstart[i] || (colstart[i+1]==colstart[i]+1 && rowindex[colstart[i]]==i)){
         ordering[order_index]=i;
         ++order_index;
         eliminated[i]=true;
      }else if(colstart[i+1]-colstart[i]>dense_threshhold){
         eliminated[i]=true; // so we don't bother with these nodes during minimum degree... (even though technically they're not eliminated yet)
         if(constrained[i]) constrained_at_end.push_back(i);
         else{
            unconstrained_at_end.push_back(i);
            // also delay its constrained neighbours if saying safe!
            if(safely_handle_dense_nodes){
               for(KKT_INT a=colstart[i]; a<colstart[i+1]; ++a){
                  KKT_INT j=rowindex[a];
                  if(constrained[j] && !eliminated[j]){
                     eliminated[j]=true;
                     constrained_at_end.push_back(j);
                  }
               }
            }
         }
      }
   }
   // Initialize quotient graph in adj, and constrained and superlist while we're at it, with what's left.
   for(KKT_INT i=0; i<n; ++i){
      if(!eliminated[i]){
         // initialize this bit of the quotient graph.
         adj[i].reserve(colstart[i+1]-colstart[i]);
         for(KKT_INT a=colstart[i]; a<colstart[i+1]; ++a){
            if(rowindex[a]!=i){ // ignore diagonal entries
               assert(rowindex[a]>=0 && rowindex[a]<n);
               if(!eliminated[rowindex[a]]){ // don't include links to removed dense rows/columns
                  adj[i].push_back(rowindex[a]);
               }
            }
         }
         // Make sure it's its own supernode (we'll properly detect supernodes later).
         superlist[i].push_back(i);
      }
   }
}

void KKTOrdering::
do_the_ordering(void)
{
   DEBUG(print_state());
   // Do initial check for supernodes in graph (with no quotient nodes to worry about yet)
   // and from there figure out costs and put what's needed in the pq.
   DEBUG(std::cerr<<"INITIAL SCAN ========================================"<<std::endl);
   initialize_ordering();
   DEBUG(print_state());
   // Then do the greedy ordering.
   while(!pq.empty()){
      // Pick a minimum cost node.
      assert(order_index<n);
      KKT_INT i=-1;
      for(;;){
         i=pq.find_min();
         if(accurate_cost[i]){
            break;
         }else{ // if this might not be minimum cost after all, update and try again
            update_cost(i);
            accurate_cost[i]=true;
            ++number_of_cost_updates[i];
         }
      }
      KKT_UINT size_i=superlist[i].size();
      DEBUG(std::cerr<<"PICKED "<<i<<" ======================================"<<std::endl);
      assert(!eliminated[i]);
      // Eliminate it. (this will also eliminate any neighbours which cause no additional fill)
      DEBUG(std::cerr<<"ELIMINATING "<<i<<" ======================================"<<std::endl);
      eliminate_node(i);
      //DEBUG(print_state());
      // Some neighbours may now be able to be compressed into supernodes.
      find_supernodes_around_eliminated_node(i);
      // Update all affected costs. (quick and dirty estimate)
      DEBUG(std::cerr<<"UPDATING COSTS ======================================"<<std::endl);
      cheap_update_cost(i, size_i);
      DEBUG(print_state());
      DEBUG(std::cerr<<std::endl);
   }
   // finish off with any dense nodes or hangers-on that need to be eliminated at the end
   for(KKT_UINT a=0; a<unconstrained_at_end.size(); ++a){
      KKT_INT i=unconstrained_at_end[a];
      DEBUG(std::cerr<<"ADDITIONAL DENSE "<<i<<std::endl;);
      ordering[order_index]=i;
      ++order_index;
   }
   for(KKT_UINT a=0; a<constrained_at_end.size(); ++a){
      KKT_INT i=constrained_at_end[a];
      DEBUG(std::cerr<<"ADDITIONAL DENSE/CONSTRAINED "<<i<<std::endl;);
      ordering[order_index]=i;
      ++order_index;
   }
   assert(order_index==n);

   KKT_INT max_updates=0, total_updates=0;
   for(KKT_INT i=0; i<n; ++i){
      total_updates+=number_of_cost_updates[i];
      if(number_of_cost_updates[i]>max_updates) max_updates=number_of_cost_updates[i];
   }
   DEBUG(std::cerr<<" max updates="<<max_updates<<", average updates="<<total_updates/(double)n<<std::endl);
}

void KKTOrdering::
initialize_ordering(void)
{
   DEBUG(std::cerr<<"scanning for supernodes"<<std::endl);
   // first scan for supernodes (so we can compact the original graph)
   for(KKT_INT i=0; i<n; ++i){
      if(eliminated[i] || superparent[i]!=-1) continue;
      // mark the neighbourhood of i (and i itself)
      marker.clear();
      marker.set(i);
      for(KKT_UINT a=0; a<adj[i].size(); ++a){
         marker.set(adj[i][a]);
      }
      // check if any of the higher numbered neighbours have the same pattern
      for(KKT_UINT a=0; a<adj[i].size(); ++a){
         KKT_INT j=adj[i][a];
         if(j<i) continue; // only bother looking at higher numbers
         if(constrained[j]!=constrained[i]) continue; // only put nodes of the same type in a supernode
         if(adj[j].size()!=adj[i].size()) continue; // only bother looking at ones with the same degree
         bool j_is_indistinguishable_from_i=true;
         for(KKT_UINT b=0; b<adj[j].size(); ++b){
            if(!marker[adj[j][b]]){
               j_is_indistinguishable_from_i=false;
               break;
            }
         }
         if(j_is_indistinguishable_from_i){
            // join j into i's supernode
            superparent[j]=i;
            assert(superlist[j].size()==1);
            superlist[i].push_back(j);
            superlist[j].clear();
            adj[j].clear(); // not needed anymore
         }
      }
      // if we added any nodes to superlist[i], eliminate self-edges from adj[i]
      if(superlist[i].size()>1){
         for(KKT_INT a=0; a<(KKT_INT)adj[i].size(); ++a){
            KKT_INT j=adj[i][a];
            if(superparent[j]==i){ // if this is one of the nodes in superlist[i]
               remove_from_list(adj[i], a);
            }
         }
         if(adj[i].size()==0){ // if it's now an isolated supernode eliminate it right away
            for(KKT_UINT a=0; a<superlist[i].size(); ++a){
               KKT_INT j=superlist[i][a];
               eliminated[j]=true;
               ordering[order_index]=j;
               ++order_index;
            }
            superlist[i].clear();
         }
      }
   }
   DEBUG(std::cerr<<"compacting adjacency lists, setting costs, adding to pq"<<std::endl);
   // make another pass through the graph, compacting adjacency lists due to supernodes,
   // setting costs and putting nodes in the pq as needed.
   for(KKT_INT i=0; i<n; ++i){
      if(eliminated[i] || superparent[i]!=-1) continue;
      bool can_order=true; // if i is constrained but has uneliminated neighbours, we'll set this to false
      for(KKT_INT a=0; a<(KKT_INT)adj[i].size(); ++a){
         KKT_INT j=adj[i][a];
         if(superparent[j]!=-1){ // if this neighbour isn't the root of a supernode
            // remove it from the adjacency list (note: the supernode's root must also be in the adjacency list)
            remove_from_list(adj[i], a);
         }else if(constrained[i] && !constrained[j]){
            can_order=false;
         }
      }
      initialize_cost(i);
      accurate_cost[i]=true;
      ++number_of_cost_updates[i];
      if(can_order){
         DEBUG(std::cerr<<"Adding "<<i<<" to pq"<<std::endl);
         pq.add(i);
      }
   }
}

void KKTOrdering::
eliminate_node(KKT_INT i)
{
   // Take it from the pq.
   assert(pq.contains(i));
   pq.remove(i);
   assert(!pq.contains(i));
   // Order all its constituent nodes.
   assert(superparent[i]==-1); // this should be the root of a supernode if it was in the pq!
   for(KKT_UINT a=0; a<superlist[i].size(); ++a){
      KKT_INT j=superlist[i][a];
      assert(!eliminated[j]);
      eliminated[j]=true;
      ordering[order_index]=j;
      ++order_index;
   }
   superlist[i].clear(); // kill the superlist - now irrelevant
   // Merge neighbouring elements (quotient nodes) into this one, update adj.
   std::vector<KKT_INT> newadj;
   newadj.reserve(adj[i].size());
   marker.clear();
   marker.set(i); // make sure we don't add a self-edge to i by marking it right away
   for(KKT_UINT a=0; a<adj[i].size(); ++a){
      KKT_INT j=find_super_root(adj[i][a]);
      if(!eliminated[j]){
         if(!marker[j]){
            newadj.push_back(j);
            marker.set(j);
         }
      }else{
         if(j==i) continue; // ignore self-edges
         superparent[j]=i; // chain the new quotient node to this one
         // Add unique neighbours of j to newadj.
         for(KKT_UINT b=0; b<adj[j].size(); ++b){
            KKT_INT k=adj[j][b];
            if(eliminated[k]) continue; // already in this quotient supernode
            k=find_super_root(k);
            if(!marker[k]){
               newadj.push_back(k);
               marker.set(k);
            }
         }
         // Empty out j since i is the new root.
         adj[j].clear();
      }
   }
   // Give the new adjacency list to i.
   swap(adj[i], newadj);
   // Prune edges from orderable neighbours of i to other neighbours of i
   // (since all neighbours are now implicitly connected via i).
   // If this leaves a neighbour only pointing to i, eliminate it too since no fill is caused
   // (this is the mass elimination technique).
   for(KKT_INT a=0; a<(KKT_INT)adj[i].size(); ++a){
      KKT_INT j=adj[i][a];
      assert(superparent[j]==-1);
      if(pq.contains(j)){
         bool j_can_be_mass_eliminated=true;
         for(KKT_INT b=0; b<(KKT_INT)adj[j].size(); ++b){
            KKT_INT k=find_super_root(adj[j][b]);
            if(k==i) continue; // ignore edge to eliminated node i
            if(marker[k]){
               remove_from_list(adj[j], b);
            }else{
               j_can_be_mass_eliminated=false;
            }
         }
         if(j_can_be_mass_eliminated){
            for(KKT_UINT b=0; b<superlist[j].size(); ++b){
               KKT_INT k=superlist[j][b];
               assert(!eliminated[k]);
               eliminated[k]=true;
               superparent[k]=i; // merge it with quotient node i
               ordering[order_index]=k;
               ++order_index;
            }
            superlist[j].clear(); // kill the superlist - now irrelevant
            adj[j].clear(); // no need to keep this either
            pq.remove(j); // and take it off the pq
            remove_from_list(adj[i], a); // and remove from i's adj list
         }
      }
   }
   // And finally check if eliminating i (and no-fill neighbours) releases any constrained nodes to be ordered.
   for(KKT_UINT a=0; a<adj[i].size(); ++a){
      KKT_INT j=adj[i][a];
      assert(!eliminated[j]);
      if(constrained[j] && !pq.contains(j)){
         DEBUG(std::cerr<<" -> checking if we can add "<<j<<" to pq"<<std::endl);
         check_constrained_node(j);
      }
   }
}

void KKTOrdering::
check_constrained_node(KKT_INT j)
{
   assert(constrained[j]);
   assert(!eliminated[j]);
   assert(!pq.contains(j));
   assert(superparent[j]==-1);
   marker.clear(); // also eliminate some redundant entries while we're at it
   for(KKT_INT a=0; a<(KKT_INT)adj[j].size(); ++a){
      KKT_INT k=find_super_root(adj[j][a]);
      if(marker[k]){
         remove_from_list(adj[j], a);
         continue;
      }
      marker.set(k);
      if(!constrained[k] && !eliminated[k]){
         DEBUG(std::cerr<<" CAN'T ADD "<<j<<" to pq because unconstrained neighbour "<<k<<" is not eliminated."<<std::endl);
         return; // j still needs to wait for k to be eliminated
      }
   }
   // if we made it this far, j can be considered for ordering
   DEBUG(std::cerr<<" newly adding constrained node "<<j<<" to pq"<<std::endl);
   pq.add(j);
}

void KKTOrdering::
find_supernodes_around_eliminated_node(KKT_INT i)
{
   HashTable hash_to_node(adj[i].size());
   std::vector<KKT_INT> possiblematches;
   for(KKT_INT a=0; a<(KKT_INT)adj[i].size(); ++a){
      KKT_INT j=adj[i][a];
      // remove redundant entries from j's adj list, constructing a simple hash of j as we go
      marker.clear();
      KKT_INT jhash=0;
      for(KKT_INT b=0; b<(KKT_INT)adj[j].size(); ++b){
         KKT_INT k=find_super_root(adj[j][b]);
         if(!marker[k]){ // if this is something we haven't seen before
            marker.set(k);
            jhash+=k;
         }else{ // otherwise get rid of it (redundant)
            remove_from_list(adj[j], b);
         }
      }
      if(constrained[j]) jhash=-jhash; // add in a sign bit to the hash to distinguish constrained and unconstrained nodes
      // check against previously seen nodes with the same hash
      KKT_INT matched_node=-1;
      hash_to_node.find(jhash, possiblematches);
      for(KKT_UINT c=0; c<possiblematches.size(); ++c){
         KKT_INT k=possiblematches[c];
         if(adj[k].size()<adj[j].size()) continue; // if k has less unique neighbours, it can't be indistinguishable
                                                   // (note: if k has more entries in adj[k], it may still have
                                                   // exactly the same neighbours - just adj[k] has some redundancy
                                                   // due to an intermediate node being supermerged)
         if(constrained[k]!=constrained[j]) continue; // we don't merge a constrained and unconstrained node together
         // test k against j's adj marker
         bool j_is_indistinguishable_from_k=true;
         for(KKT_UINT b=0; b<adj[k].size(); ++b){
            KKT_INT l=find_super_root(adj[k][b]);
            if(!marker[l]){
               j_is_indistinguishable_from_k=false;
               break;
            }
         }
         // join j up to k if indistinguishable
         if(j_is_indistinguishable_from_k){
            matched_node=k;
            break;
         }
      }
      // if we could do a merge
      if(matched_node>=0){
         // merge j into matched_node
         superparent[j]=matched_node;
         /*
         if(superlist[j].size()>superlist[matched_node].size()){
            swap(superlist[j], superlist[matched_node]);
         }
         */
         for(KKT_UINT b=0; b<superlist[j].size(); ++b){
            superlist[matched_node].push_back(superlist[j][b]);
         }
         superlist[j].clear();
         // we'll use adj[j] for matched_node, since it's more up-to-date
         swap(adj[j], adj[matched_node]);
         adj[j].clear(); // not needed any more
         // and finally, remove redundant entry j in i's adj list
         remove_from_list(adj[i], a);
         // and take j out of the priority queue (matched_node will stay in the pq if needed)
         assert(pq.contains(j)==pq.contains(matched_node));
         if(pq.contains(j)) pq.remove(j);
      }else{ // if j is different from all the previous nodes, make a record of its hash value
         hash_to_node.add(jhash, j);
      }
   }
}

KKT_INT KKTOrdering::
find_super_root(KKT_INT &j)
{
   if(superparent[j]==-1){
      return j;
   }else{
      // first find root
      KKT_INT r=superparent[j], next_up;
      for(;;){
         next_up=superparent[r];
         if(next_up==-1) break;
         r=next_up;
      }
      // compress path
      while(j!=r){
         next_up=superparent[j];
         superparent[j]=r;
         j=next_up;
      }
      return r;
   }
}

void KKTOrdering::
print_state(void)
{
return;
   std::cerr<<"------------------------------------------------------------------"<<std::endl;
   std::cerr<<"ordered "<<order_index<<" / "<<n<<std::endl;
   for(KKT_INT i=0; i<order_index; ++i){
      std::cerr<<" "<<ordering[i];
   }
   std::cerr<<std::endl;
   for(KKT_INT i=0; i<n; ++i){
      std::cerr<<"Node "<<i<<": ";
      if(constrained[i]) std::cerr<<"constrained,     ";
      else               std::cerr<<"NOT constrained, ";
      if(eliminated[i]) std::cerr<<"eliminated,     ";
      else              std::cerr<<"NOT eliminated, ";
      if(pq.contains(i)) std::cerr<<"in pq"<<std::endl;
      else               std::cerr<<"NOT in pq"<<std::endl;
      std::cerr<<"  cost = "<<cost[i]<<std::endl;
      std::cerr<<"  superparent = "<<superparent[i]<<std::endl;
      if(superparent[i]!=-1){
         KKT_INT j=i;
         j=find_super_root(j);
         bool found=false;
         for(KKT_UINT a=0; a<superlist[j].size(); ++a){
            if(superlist[j][a]==i){
               found=true;
               break;
            }
         }
         if(found){
            std::cerr<<"    found in super root "<<j<<"'s superlist"<<std::endl;
         }else{
            std::cerr<<"    NOT in super root "<<j<<"'s superlist"<<std::endl;
         }
      }
      std::cerr<<"  superlist =";
      for(KKT_UINT a=0; a<superlist[i].size(); ++a){
         std::cerr<<" "<<superlist[i][a];
      }
      std::cerr<<std::endl;
      std::cerr<<"  adj =";
      for(KKT_UINT a=0; a<adj[i].size(); ++a){
         std::cerr<<" "<<adj[i][a];
      }
      std::cerr<<std::endl;
   }
}

//=================== subclass (specific cost metric) ====================

struct KKTOrdering_ExactExternalDegree: public KKTOrdering
{
   KKTOrdering_ExactExternalDegree(KKT_INT n_, const KKT_INT *colstart, const KKT_INT *rowindex, const char *constrained_, int safely_handle_dense_nodes, KKT_INT *ordering_)
      : KKTOrdering(n_, colstart, rowindex, constrained_, safely_handle_dense_nodes, ordering_)
   {}

   void initialize_cost(KKT_INT i)
   {
      cost[i]=0;
      for(KKT_UINT a=0; a<adj[i].size(); ++a){
         KKT_INT j=adj[i][a];
         assert(superparent[j]==-1);
         cost[i]+=superlist[j].size();
      }
   }

   void update_cost(KKT_INT j)
   {
      assert(superparent[j]==-1);
      assert(pq.contains(j));
      KKT_UINT newcost=0;
      marker.clear();
      marker.set(j);
      for(KKT_INT b=0; b<(KKT_INT)adj[j].size(); ++b){
         KKT_INT k=find_super_root(adj[j][b]);
         if(!eliminated[k]){
            if(!marker[k]){
               newcost+=superlist[k].size();
               marker.set(k);
            }else{
               remove_from_list(adj[j], b);
            }
         }else{
            // add in the contribution from quotient node k
            for(KKT_UINT c=0; c<adj[k].size(); ++c){
               KKT_INT l=find_super_root(adj[k][c]);
               assert(!eliminated[l]);
               if(!marker[l]){
                  newcost+=superlist[l].size();
                  marker.set(l);
               }
            }
         }
      }
      pq.modify_key(j, newcost); // note: we only calculate new cost for variables in the priority queue
   }
   
   void cheap_update_cost(KKT_INT eliminated_node, KKT_UINT previous_size)
   {
      for(KKT_UINT a=0; a<adj[eliminated_node].size(); ++a){
         KKT_INT j=adj[eliminated_node][a];
         assert(!eliminated[j]);
         if(pq.contains(j)){
            if(cost[j]<=previous_size){
               update_cost(j);
               accurate_cost[j]=true;
               ++number_of_cost_updates[j];
            }else{
               pq.modify_key(j, cost[j]-previous_size);
               accurate_cost[j]=false;
            }
         }
      }
   }
};

//============================= the main function ============================

void KKT_minimum_degree(KKT_INT n,
                        const KKT_INT *colstart,
                        const KKT_INT *rowindex,
                        const char *constrained,
                        int safely_handle_dense_nodes,
                        KKT_INT *ordering)
{
   assert(n>0);
   KKTOrdering_ExactExternalDegree orderer(n, colstart, rowindex, constrained, safely_handle_dense_nodes, ordering);
   orderer.do_the_ordering();
}

