#include "kktdirect.h"
#include "kktsupport.h"

// uplooking method ==========================================================
int KKT_numerically_factor_uplooking(KKT_INT n,
                                     const KKT_INT *colstart,
                                     const KKT_INT *rowindex,
                                     const double *value,
                                     const KKT_INT *postordering,
                                     const KKT_INT *inverseordering,
                                     const KKT_INT *etreeparent,
                                     const KKT_INT *factorcolstart,
                                     KKT_INT *factorrowindex,
                                     double *factorvalue,
                                     double *diagvalue)
{
   if(n<1) return 1;
   int pivots_are_all_fine=1; // zero this flag if we encounter a dangerously small or zero pivot
   // calculate lower factor (and diagonal) one row at a time, but storing it columnwise.
   std::vector<KKT_INT> factorcolend(n);
   for(KKT_INT a=0; a<n; ++a) factorcolend[a]=factorcolstart[a];
   Marker marker(n);
   std::vector<KKT_INT> factorrownonzeros;
   factorrownonzeros.reserve(n);
   std::vector<KKT_INT> stack;
   stack.reserve(n);
   std::vector<double> y(n,0);
   for(KKT_INT a=0; a<n; ++a){
      KKT_INT i=postordering[a];
      // Get a topological ordering of the reachable set in L(1:a-1,1:a-1) from A(post(1:a-1),post(a))
      // (stored *in reverse* in factorrownonzeros) by a sequence of traversals of the elimination tree.
      // Also when we find the diagonal entry in A (if there is one) record it for later
      // And finally, scatter numerical enries of this column of A in vector y
      factorrownonzeros.resize(0);
      marker.clear();
      marker.set(a);
      double diagii=0; // where we record the A(i,i) diagonal entry
      for(KKT_INT b=colstart[i]; b<colstart[i+1]; ++b){
         KKT_INT j=rowindex[b];
         KKT_INT c=inverseordering[j];
         if(c>a) continue;
         if(c==a){
            diagii=value[b];
            continue;
         }
         assert(y[c]==0);
         y[c]=value[b];
         assert(stack.empty());
         while(!marker[c]){
            stack.push_back(c);
            marker.set(c);
            j=etreeparent[j];
            c=inverseordering[j];
            assert(c<=a);
         }
         while(!stack.empty()){
            factorrownonzeros.push_back(stack.back());
            stack.pop_back();
         }
      }
      // now figure out the numerical values of solving L(1:a-1,1:a-1)*y=A(post(1:a-1),post(a))
      // scattered in the dense vector y (though hopefully postordering means not *too* scattered!)
      // (note we start with y containing A(post(1:a-1),post(a)) from the previous loop)
      for(KKT_INT p=(KKT_INT)factorrownonzeros.size()-1; p>=0; --p){ // loop through eventual nonzeros of y in topological order
         KKT_INT c=factorrownonzeros[p];
         assert(c<a);
         double yc=y[c];
         assert(factorcolend[c]<=factorcolstart[c+1]);
         for(KKT_INT q=factorcolstart[c]; q<factorcolend[c]; ++q){
            KKT_INT d=factorrowindex[q];
            assert(d>=0 && d<a);
            y[d]-=factorvalue[q]*yc;
         }
      }
      // add the nonzeros of y, divided by appropriate diagonal, as next row of L (and clean up y as we go)
      // and accumulate dot-product L(:,a)*y for use in computing new diagonal entry
      double dot=0;
      double maxterm=0; // also keep track of biggest term in dot-product for comparison purposes later
      for(KKT_INT p=(KKT_INT)factorrownonzeros.size()-1; p>=0; --p){
         KKT_INT c=factorrownonzeros[p];
         assert(c<a);
         KKT_INT q=factorcolend[c];
         ++factorcolend[c];
         assert(factorcolend[c]<=factorcolstart[c+1]);
         factorrowindex[q]=a;
         if(diagvalue[c]!=0){
            factorvalue[q]=y[c]/diagvalue[c];
            maxterm=std::max(maxterm, std::fabs(factorvalue[q]));
            dot+=factorvalue[q]*y[c];
         }else{
            factorvalue[q]=0; // avoid the divide by zero to stop Inf or NaN appearing
         }
         y[c]=0; // clean up
      }
      // finally the diagonal entry
      diagvalue[a]=diagii-dot;
      // check for breakdown
      if(std::fabs(diagvalue[a])<=1e-12*(std::fabs(diagii)+maxterm)){
         pivots_are_all_fine=0; // flag this to the user so they know we're in dangerous territory
         // then shift this diagonal entry away from zero if we think it probably should be nonzero
         // (Note this leaves a zero if diagii is zero and dot had almost perfect cancellation)
         diagvalue[a]=1e-8*diagii;
      }
   }
   return pivots_are_all_fine;
}

// solve with point factor ===================================================
void KKT_solve_with_pointwise_factorization(KKT_INT n,
                                            const KKT_INT *postordering,
                                            const KKT_INT *factorcolstart,
                                            const KKT_INT *factorrowindex,
                                            const double *factorvalue,
                                            const double *diagvalue,
                                            const double *righthandside,
                                            double *solution)
{
   // first permute right-hand-side to chosen ordering, storing in y
   std::vector<double> y(n);
   for(KKT_INT i=0; i<n; ++i){
      y[i]=righthandside[postordering[i]];
   }
   // then solve L*ynew=y in-place
   for(KKT_INT i=0; i<n; ++i){
      double yi=y[i];
      for(KKT_INT j=factorcolstart[i]; j<factorcolstart[i+1]; ++j){
         y[factorrowindex[j]]-=factorvalue[j]*yi;
      }
   }
   // then divide by D in-place, y:=inv(D)*y
   for(KKT_INT i=0; i<n; ++i){
      if(diagvalue[i]!=0){
         y[i]/=diagvalue[i];
      }else{
         y[i]=0;
      }
   }
   // then solve L^T*ynew=y in-place
   for(KKT_INT i=n-1; i>=0; --i){
      for(KKT_INT j=factorcolstart[i]; j<factorcolstart[i+1]; ++j){
         y[i]-=factorvalue[j]*y[factorrowindex[j]];
      }
   }
   // and repermute y to get solution
   for(KKT_INT i=0; i<n; ++i){
      solution[postordering[i]]=y[i];
   }
}

//================= numerical factorization (signed Cholesky, point) ==============
// returns 0 if there was a problem (zero or very small pivot) and nonzero otherwise
int KKT_numerically_factor_signed_cholesky(KKT_INT n,
                                           const KKT_INT *colstart,
                                           const KKT_INT *rowindex,
                                           const double *value,
                                           const char *constrained,
                                           const KKT_INT *postordering,
                                           const KKT_INT *inverseordering,
                                           const KKT_INT *etreeparent,
                                           const KKT_INT *factorcolstart,
                                           KKT_INT *factorrowindex,
                                           double *factorvalue)
{
   if(n<1) return 1;
   int pivots_are_all_fine=1; // zero this flag if we encounter a dangerously small or zero pivot
   // calculate lower factor (and diagonal) one row at a time, but storing it columnwise.
   std::vector<KKT_INT> factorcolend(n);
   for(KKT_INT a=0; a<n; ++a) factorcolend[a]=factorcolstart[a]+1; // skip over diagonal entries
   Marker marker(n);
   std::vector<KKT_INT> factorrownonzeros;
   factorrownonzeros.reserve(n);
   std::vector<KKT_INT> stack;
   stack.reserve(n);
   std::vector<double> y(n,0);
   for(KKT_INT a=0; a<n; ++a){
      KKT_INT i=postordering[a];
      // Get a topological ordering of the reachable set in L(1:a-1,1:a-1) from A(post(1:a-1),post(a))
      // (stored *in reverse* in factorrownonzeros) by a sequence of traversals of the elimination tree.
      // Also when we find the diagonal entry in A (if there is one) record it for later
      // And finally, scatter numerical enries of this column of A in vector y
      factorrownonzeros.resize(0);
      marker.clear();
      marker.set(a);
      double diagii=0; // where we record the A(i,i) diagonal entry
      for(KKT_INT b=colstart[i]; b<colstart[i+1]; ++b){
         KKT_INT j=rowindex[b];
         KKT_INT c=inverseordering[j];
         if(c>a) continue;
         if(c==a){
            diagii=value[b];
            continue;
         }
         assert(y[c]==0);
         y[c]=value[b];
         assert(stack.empty());
         while(!marker[c]){
            stack.push_back(c);
            marker.set(c);
            j=etreeparent[j];
            c=inverseordering[j];
            assert(c<=a);
         }
         while(!stack.empty()){
            factorrownonzeros.push_back(stack.back());
            stack.pop_back();
         }
      }
      // now figure out the numerical values of solving L(1:a-1,1:a-1)*y=A(post(1:a-1),post(a))
      // scattered in the dense vector y (though hopefully postordering means not *too* scattered!)
      // (note we start with y containing A(post(1:a-1),post(a)) from the previous loop)
      for(KKT_INT p=(KKT_INT)factorrownonzeros.size()-1; p>=0; --p){ // loop through eventual nonzeros of y in topological order
         KKT_INT c=factorrownonzeros[p];
         assert(c<a);
         double diagc=factorvalue[factorcolstart[c]];
         double yc=(diagc!=0 ? y[c]/factorvalue[factorcolstart[c]] : 0);
         y[c]=yc;
         assert(factorcolend[c]<=factorcolstart[c+1]);
         for(KKT_INT q=factorcolstart[c]+1; q<factorcolend[c]; ++q){
            KKT_INT d=factorrowindex[q];
            assert(d>=0 && d<a);
            y[d]-=factorvalue[q]*yc;
         }
      }
      // add the nonzeros of y, with signs corrected appropriately, as next row of L, cleaning up y as we go
      // and accumulating dot-product L(:,a)*y for use in computing new diagonal entry
      double dot=0;
      double maxterm=0; // also keep track of biggest term in dot-product for comparison purposes later
      for(KKT_INT p=(KKT_INT)factorrownonzeros.size()-1; p>=0; --p){
         KKT_INT c=factorrownonzeros[p];
         assert(c<a);
         KKT_INT q=factorcolend[c];
         ++factorcolend[c];
         assert(factorcolend[c]<=factorcolstart[c+1]);
         factorrowindex[q]=a;
         if(constrained[postordering[c]]){
            factorvalue[q]=-y[c];
         }else{
            factorvalue[q]=y[c];
         }
         maxterm=std::max(maxterm, std::fabs(factorvalue[q]));
         dot+=factorvalue[q]*y[c];
         y[c]=0; // clean up
      }
      // finally the diagonal entry
      double diaga=diagii-dot;
      // check for breakdown
      if((constrained[postordering[a]] && diaga>=-1e-12*(std::fabs(diagii)+maxterm))
      || (!constrained[postordering[a]] && diaga<=1e-12*(std::fabs(diagii)+maxterm))){
         pivots_are_all_fine=0; // flag this to the user so they know we're in dangerous territory
         // then shift this diagonal entry away from zero if we think it probably should be nonzero
         // (Note this leaves a zero if diagii is zero and dot had almost perfect cancellation)
         diaga=1e-8*diagii;
      }
      factorrowindex[factorcolstart[a]]=a;
      factorvalue[factorcolstart[a]]=std::sqrt(std::fabs(diaga));
   }
   return pivots_are_all_fine;
}

//============== numerical solution (signed Cholesky, point) ===================
void KKT_solve_with_signed_cholesky(KKT_INT n,
                                    const char *constrained,
                                    const KKT_INT *postordering,
                                    const KKT_INT *factorcolstart,
                                    const KKT_INT *factorrowindex,
                                    const double *factorvalue,
                                    const double *righthandside,
                                    double *solution)
{
   // first permute right-hand-side to chosen ordering, storing in y
   std::vector<double> y(n);
   for(KKT_INT i=0; i<n; ++i){
      y[i]=righthandside[postordering[i]];
   }
   // then solve L*ynew=y in-place
   for(KKT_INT i=0; i<n; ++i){
      if(factorvalue[factorcolstart[i]]){
         y[i]/=factorvalue[factorcolstart[i]];
      }else{
         y[i]=0;
      }
      double yi=y[i];
      y[i]=yi;
      for(KKT_INT j=factorcolstart[i]+1; j<factorcolstart[i+1]; ++j){
         y[factorrowindex[j]]-=factorvalue[j]*yi;
      }
   }
   // then handle signs
   for(KKT_INT i=0; i<n; ++i){
      if(constrained[postordering[i]]) y[i]=-y[i];
   }
   // then solve L^T*ynew=y in-place
   for(KKT_INT i=n-1; i>=0; --i){
      for(KKT_INT j=factorcolstart[i]+1; j<factorcolstart[i+1]; ++j){
         y[i]-=factorvalue[j]*y[factorrowindex[j]];
      }
      if(factorvalue[factorcolstart[i]]){
         y[i]/=factorvalue[factorcolstart[i]];
      }else{
         y[i]=0;
      }
   }
   // and repermute y to get solution
   for(KKT_INT i=0; i<n; ++i){
      solution[postordering[i]]=y[i];
   }
}

