#include "mex.h"
#include "kktdirect.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
   // check argument counts
   if(nrhs<2 || nrhs>3) mexErrMsgTxt("Wrong number of input arguments");
   if(nlhs==0) return;
   if(nlhs>1) mexErrMsgTxt("Too many output arguments (should just be solution of linear system)");

   // check on right hand side of linear system
   KKT_INT n=mxGetM(prhs[1]);
   if(n==0) mexErrMsgTxt("Empty right hand side");
   KKT_INT nrhs_system=mxGetN(prhs[1]); // note: not the same as nrhs in mexFunction
   if(nrhs_system==0) mexErrMsgTxt("Empty right hand side");
   if(!mxIsDouble(prhs[1])) mexErrMsgTxt("Only can solve with real, double-precision, dense right-hand-sides");

   // check on matrix from linear system
   if(mxGetNumberOfDimensions(prhs[0])!=2) mexErrMsgTxt("Input matrix is not two dimensional");
   if(mxGetM(prhs[0])!=mxGetN(prhs[0])) mexErrMsgTxt("Input matrix is not square");
   if(mxGetM(prhs[0])!=n) mexErrMsgTxt("Input right hand side is different size than matrix!");
   if(!mxIsSparse(prhs[0])) mexErrMsgTxt("Input matrix is not sparse");
   // ideally we would check symmetry too, but for now we'll let the m-file wrapper handle that test

   // check if we should use pointwise uplooking version instead of supernodes
   bool use_supernodes=true;
   if(nrhs==3){
      if(!mxIsDouble(prhs[2])) mexErrMsgTxt("Pass in options as a dense array of doubles");
      if(mxGetNumberOfElements(prhs[2])>0){
         if(mxGetPr(prhs[2])[0]==0){
            use_supernodes=false;
         }
      }
   }

   // Start the work ======================================================================================

   // order the matrix
   KKT_INT *colstart=mxGetJc(prhs[0]);
   KKT_INT *rowindex=mxGetIr(prhs[0]);
   const double *value=mxGetPr(prhs[0]);
   char *constrained=(char*)mxCalloc(n, sizeof(char));
   KKT_detect_constraints(n, colstart, rowindex, value,
                          constrained);

   KKT_INT *order=(KKT_INT*)mxCalloc(n, sizeof(KKT_INT));
   KKT_minimum_degree(n, colstart, rowindex, constrained, 1, 1,
                      order);
   if(!KKT_check_constraints_satisified_by_ordering(n, colstart, rowindex, constrained, order)){
      mexWarnMsgTxt("error in minimum degree: doesn't satisfy constraints\n");
   }

   // compute a postordering and elimination tree
   KKT_INT *postordering=(KKT_INT*)mxCalloc(n, sizeof(KKT_INT));
   KKT_INT *inverseordering=(KKT_INT*)mxCalloc(n, sizeof(KKT_INT));
   KKT_INT *etreeparent=(KKT_INT*)mxCalloc(n, sizeof(KKT_INT));
   KKT_postorder(n, colstart, rowindex, order,
                 postordering, inverseordering, etreeparent);
   if(!KKT_check_constraints_satisified_by_ordering(n, colstart, rowindex, constrained, postordering)){
      mexWarnMsgTxt("error in postordering: doesn't satisfy constraints\n");
   }
   mxFree(order);
   mxFree(constrained);

   // now do the factorization and solve!

   if(use_supernodes){
      // first the symbolic factorization
      KKT_INT numsupernodes;
      KKT_INT *supernodestart=(KKT_INT*)mxCalloc(n+1, sizeof(KKT_INT));
      KKT_INT *supernodemap=(KKT_INT*)mxCalloc(n, sizeof(KKT_INT));
      KKT_INT *supercolstart=(KKT_INT*)mxCalloc(n+1, sizeof(KKT_INT));
      KKT_INT *factorcolstart=(KKT_INT*)mxCalloc(n+1, sizeof(KKT_INT));
      KKT_INT *diagblockstart=(KKT_INT*)mxCalloc(n+1, sizeof(KKT_INT)); 
      KKT_INT max_supernode_size=64;
      KKT_symbolically_factor_supernodal_initial(n, colstart, rowindex, postordering, inverseordering, etreeparent, max_supernode_size,
                        &numsupernodes, supernodestart, supernodemap, supercolstart, factorcolstart, diagblockstart);
      KKT_INT *superrowindex=(KKT_INT*)mxCalloc(supercolstart[numsupernodes], sizeof(KKT_INT));
      KKT_symbolically_factor_supernodal_final(n, colstart, rowindex, postordering, inverseordering, etreeparent,
                        numsupernodes, supernodestart, supernodemap, supercolstart,
                        superrowindex);

      // then the numerical factorization
      double *factorvalue=(double*)mxCalloc(factorcolstart[numsupernodes], sizeof(double));
      double *diagvalue=(double*)mxCalloc(diagblockstart[numsupernodes], sizeof(double));
      KKT_INT *diagpiv=(KKT_INT*)mxCalloc(n, sizeof(KKT_INT));
      int flag=KKT_numerically_factor_leftlooking_supernodal(n, colstart, rowindex, value, postordering,
                        inverseordering, etreeparent, numsupernodes, supernodestart, supernodemap, supercolstart,
                        superrowindex, factorcolstart, factorvalue, diagblockstart, diagvalue, diagpiv);
      if(flag==0){
         mexWarnMsgTxt("Supernodal left-looking factorization encountered a small or zero pivot");
      }

      // and finally the solve
      const double *righthandside=mxGetPr(prhs[1]);
      plhs[0]=mxCreateDoubleMatrix(n, nrhs_system, mxREAL);
      double *solution=mxGetPr(plhs[0]);
      KKT_solve_with_supernodal_factorization(n, postordering, numsupernodes, supernodestart, supernodemap,
                                              supercolstart, superrowindex, factorcolstart, factorvalue,
                                              diagblockstart, diagvalue, diagpiv,
                                              nrhs_system, righthandside,
                                              solution);
   
      // and clean up after ourselves
      mxFree(supernodestart);
      mxFree(supernodemap);
      mxFree(supercolstart);
      mxFree(factorcolstart);
      mxFree(diagblockstart);
      mxFree(superrowindex);
      mxFree(factorvalue);
      mxFree(diagvalue);
      mxFree(diagpiv);
   }
   else // if (!use_supernodes) 
   {
      // first the symbolic factorization
      KKT_INT *factorcolstart=(KKT_INT*)mxCalloc(n+1, sizeof(KKT_INT));
      KKT_symbolically_factor_uplooking(n, colstart, rowindex, postordering, inverseordering, etreeparent,
                                        factorcolstart);

      // then the numerical factorization
      KKT_INT *factorrowindex=(KKT_INT*)mxCalloc(factorcolstart[n], sizeof(KKT_INT));
      double *factorvalue=(double*)mxCalloc(factorcolstart[n], sizeof(double));
      double *diagvalue=(double*)mxCalloc(n, sizeof(double));
      int flag=KKT_numerically_factor_uplooking(n, colstart, rowindex, value, postordering,
                                                inverseordering, etreeparent, factorcolstart,
                                                factorrowindex, factorvalue, diagvalue);
      if(flag==0){
         mexWarnMsgTxt("Pointwise up-looking factorization encountered a small or zero pivot");
      }

      // and finally the solve
      const double *righthandside=mxGetPr(prhs[1]);
      plhs[0]=mxCreateDoubleMatrix(n, nrhs_system, mxREAL);
      double *solution=mxGetPr(plhs[0]);
      for(KKT_INT i=0; i<nrhs_system; ++i){
          KKT_solve_with_pointwise_factorization(n, postordering, factorcolstart, factorrowindex,
                                                 factorvalue, diagvalue, righthandside+i*n,
                                                 solution+i*n);
      }
   
      // and clean up after ourselves
      mxFree(factorcolstart);
      mxFree(factorrowindex);
      mxFree(factorvalue);
      mxFree(diagvalue);
   }
   
   mxFree(postordering);
   mxFree(etreeparent);
   mxFree(inverseordering);
}

