#include "kktdirect.h"
#include "kktsupport.h"
#include "dense.h"

// leftlooking method (supernodal) ===========================================
// returns 0 if there was a problem and nonzero otherwise
int KKT_numerically_factor_leftlooking_supernodal(KKT_INT n,
                                                  const KKT_INT *colstart,
                                                  const KKT_INT *rowindex,
                                                  const double *value,
                                                  const KKT_INT *postordering,
                                                  const KKT_INT *inverseordering,
                                                  const KKT_INT *etreeparent,
                                                  // description of supernodes
                                                  KKT_INT numsupernodes,
                                                  const KKT_INT *supernodestart,
                                                  const KKT_INT *supernodemap,
                                                  // description of structure of factor (one supercolumn at a time)
                                                  const KKT_INT *supercolstart,
                                                  const KKT_INT *superrowindex,
                                                  // description of values in factor
                                                  const KKT_INT *factorcolstart,
                                                  double *factorvalue, // written to in output
                                                  // description of values in diagonal
                                                  const KKT_INT *diagblockstart,
                                                  double *diagvalue,
                                                  // extra stuff for pivoting within diagonal blocks
                                                  KKT_INT *diagpiv)
{
   if(n<1) return 1;
   std::fill_n(factorvalue, factorcolstart[numsupernodes], (double)0); // zero out factor to begin
   int pivots_are_all_fine=1; // zero this flag if we encounter a dangerously small or zero pivot
   std::vector<KKT_INT> supercolpointer(numsupernodes); // keep track of where we are in structure of each column (progress row by row)
   std::copy(supercolstart, supercolstart+numsupernodes, supercolpointer.begin());
   std::vector<KKT_INT> supercolnextpointer(numsupernodes);
   std::vector<KKT_INT> factorrownonzeros; factorrownonzeros.reserve(numsupernodes);
   std::vector<KKT_INT> colstart_within_row; colstart_within_row.reserve(numsupernodes+1);
   std::vector<double> scaledrowvalue; scaledrowvalue.reserve(n);
   std::vector<double> diag_unfactored(diagblockstart[numsupernodes], 0); // where we store unfactored diagonal blocks
   std::vector<double> temp(sqr(n/numsupernodes)); // temporary storage used in various places...
   Marker marker(n), supermarker(numsupernodes);
   std::vector<KKT_INT> colmap(n); // for column of L under construction, maps nonzero rows to location in sparse structure
   for(KKT_INT s=0; s<numsupernodes; ++s){
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      // scatter the diagonal block of the original matrix into diag_unfactored
      for(KKT_INT a=supernodestart[s]; a<supernodestart[s+1]; ++a){
         KKT_INT i=postordering[a];
         for(KKT_INT p=colstart[i]; p<colstart[i+1]; ++p){
            KKT_INT j=rowindex[p];
            KKT_INT b=inverseordering[j];
            if(supernodemap[b]==s){
               KKT_INT location=diagblockstart[s] + (b-supernodestart[s])+supersize*(a-supernodestart[s]);
               diag_unfactored[location]=value[p];
            }
         }
      }
      // find supercolumn structure of this superrow using etree
      factorrownonzeros.resize(0);
      supermarker.clear();
      supermarker.set(s); // don't include diagonal block entries
      for(KKT_INT a=supernodestart[s]; a<supernodestart[s+1]; ++a){
         marker.clear();
         marker.set(a);
         KKT_INT i=postordering[a];
         for(KKT_INT p=colstart[i]; p<colstart[i+1]; ++p){
            KKT_INT j=rowindex[p];
            KKT_INT b=inverseordering[j];
            if(b>=a) continue;
            KKT_INT t=supernodemap[b];
            while(!marker[b]){
               marker.set(b);
               if(!supermarker[t]){
                  factorrownonzeros.push_back(t);
                  supermarker.set(t);
               }
               j=etreeparent[j];
               b=inverseordering[j];
               t=supernodemap[b];
            }
         }
      }
      // establish sparsity structure of this superrow (for each nonempty supercolumn, which rows are nonzero)
      colstart_within_row.resize(factorrownonzeros.size()+1);
      colstart_within_row[0]=0;
      KKT_INT nnz_in_superrow=0;
      for(KKT_UINT p=0; p<factorrownonzeros.size(); ++p){
         KKT_INT t=factorrownonzeros[p]; // supernode index of this column within the row
         assert(supercolpointer[t]==supercolstart[t+1] || superrowindex[supercolpointer[t]]>=supernodestart[s]);
         KKT_INT tsize=supernodestart[t+1]-supernodestart[t]; // how wide this supercolum is
         KKT_INT q;
         for(q=supercolpointer[t]; q<supercolstart[t+1] && superrowindex[q]<supernodestart[s+1]; ++q){ // look for which rows are nonempty
            assert(superrowindex[q]>=supernodestart[s]);
            nnz_in_superrow+=tsize;
         }
         supercolnextpointer[t]=q;
         colstart_within_row[p+1]=nnz_in_superrow;
      }
      // compute scaled superrow L(s,1:s-1)*D(1:s-1,1:s-1) in factorrownonzeros,
      // and subtract off inner product of superrow and scaled superrow from diag_unfactored
      scaledrowvalue.resize(nnz_in_superrow);
      for(KKT_UINT p=0; p<factorrownonzeros.size(); ++p){
         KKT_INT t=factorrownonzeros[p]; // supernode index of this column within the row
         KKT_INT tsize=supernodestart[t+1]-supernodestart[t]; // how wide supercolumn t is
         KKT_INT blocksize=colstart_within_row[p+1]-colstart_within_row[p]; // how many nonzeros in this chunk of the factor
         assert(blocksize%tsize==0);
         KKT_INT rsize=blocksize/tsize; // how many nonzero rows in this chunk
         // multiply block from factorvalue by diag_unfactored(t...) and store in scaledrowvalue
         //   factorvalue submatrix is at factorvalue+factorcolpointer_t, with dimensions rsize*tsize (stored row-wise)
         KKT_INT factorcolpointer_t=factorcolstart[t]+tsize*(supercolpointer[t]-supercolstart[t]);
         const double *factorsubmatrix=factorvalue+factorcolpointer_t;
         //   diag submatrix is at diag_unfactored+diagblockstart[t], with dimensions tsize*tsize (symmetric)
         const double *diagsubmatrix=&diag_unfactored[diagblockstart[t]];
         //   scaledrowvalue submatrix is at scaledrowvalue+colstart_within_row[p], with dimensions rsize*tsize (stored row-wise)
         double *scaledsubmatrix=&scaledrowvalue[colstart_within_row[p]];
         // Note: since factorsubmatrix and scaledsubmatrix are stored row-wise, not column-wise (as BLAS expect), this looks backwards --- but it isn't.
         BLAS::dgemm_NN(tsize, rsize, tsize, diagsubmatrix, factorsubmatrix, scaledsubmatrix);
         // multiply this part of the scaled row by the original row, storing in a temporary matrix
         temp.resize(rsize*rsize);
         BLAS::dgemm_TN(rsize, rsize, tsize, factorsubmatrix, scaledsubmatrix, &temp[0]);
         // and now subtract off that product from the D(s,s) unfactored diagonal block (a scattered/sparse update)
         for(KKT_INT q=0; q<rsize; ++q){
            KKT_INT x=superrowindex[supercolpointer[t]+q]-supernodestart[s];
            KKT_INT diagindex=diagblockstart[s]+x*supersize;
            for(KKT_INT r=0; r<rsize; ++r){
               KKT_INT y=superrowindex[supercolpointer[t]+r]-supernodestart[s];
               diag_unfactored[diagindex+y]-=temp[q*rsize+r];
            }
         }
      }
      // factor diagonal block D(s,s)
      std::copy(&diag_unfactored[diagblockstart[s]], &diag_unfactored[diagblockstart[s]]+sqr(supersize), diagvalue+diagblockstart[s]);
      KKT_INT info;
      if(temp.size()<sqr((KKT_UINT)supersize)) temp.resize(sqr(supersize)); // should be large enough for efficient blocking!
      LAPACK::factor_symmetric_indefinite(supersize, diagvalue+diagblockstart[s], diagpiv+supernodestart[s], &temp[0], temp.size(), info);
      if(info!=0){ // if there was a problem
         pivots_are_all_fine=0;
      }
      // figure out supercolumn s of factor, L(s+1:end,s) = (A(s+1:end,s)-L(s+1:end,1:s-1)*scaled_row^T)*inv(diag(s,s))
      //   first set up colmap (given a row, where the values are stored in the factor)
      for(KKT_INT p=supercolstart[s]; p<supercolstart[s+1]; ++p){
         colmap[superrowindex[p]]=factorcolstart[s]+(p-supercolstart[s])*supersize;
      }
      //   then copy entries from A into factor
      for(KKT_INT a=supernodestart[s]; a<supernodestart[s+1]; ++a){
         KKT_INT i=postordering[a];
         for(KKT_INT p=colstart[i]; p<colstart[i+1]; ++p){
            KKT_INT j=rowindex[p];
            KKT_INT b=inverseordering[j];
            if(supernodemap[b]>s){ // copy nonzeros below the diagonal superblock
               KKT_INT location=colmap[b]+(a-supernodestart[s]);
               assert(location>=factorcolstart[s] && location<factorcolstart[s+1]);
               factorvalue[location]=value[p];
            }
         }
      }
      //   and then subtract off factor*scaled_row^T update, one supercolumn at a time
      for(KKT_UINT p=0; p<factorrownonzeros.size(); ++p){
         KKT_INT t=factorrownonzeros[p]; // supernode index of this column within the row
         KKT_INT tsize=supernodestart[t+1]-supernodestart[t]; // how wide supercolumn t is
         KKT_INT blocksize=colstart_within_row[p+1]-colstart_within_row[p]; // how many nonzeros in supercolumn t of the (scaled) superrow
         KKT_INT rsize=blocksize/tsize; // how many nonzero rows in supercolumn t of the (scaled) superrow
         KKT_INT lsize=supercolstart[t+1]-supercolnextpointer[t]; // how many nonzero rows in the remainder of supercolumn t of the factor
         if(lsize==0) continue; // nothing to do here
         assert(supercolnextpointer[t]>=supercolpointer[t]);
         assert(supercolpointer[t]>=supercolstart[t]);
         assert(lsize<=n-supernodestart[s+1]);
         KKT_INT lstart=factorcolstart[t]+tsize*(supercolnextpointer[t]-supercolstart[t]); // where the nonzeros of the remainder of supercolumn in t begin
         assert(lstart<factorcolstart[t+1]);
         temp.resize(rsize*lsize);
         BLAS::dgemm_TN(rsize, lsize, tsize, &scaledrowvalue[colstart_within_row[p]], factorvalue+lstart, &temp[0]);
         // now scatter that product into the appropriate entries of supercolumn s
         KKT_INT tempindex=0;
         for(KKT_INT q=supercolnextpointer[t]; q<supercolstart[t+1]; ++q){
            KKT_INT y=colmap[superrowindex[q]];
            for(KKT_INT r=supercolpointer[t]; r<supercolnextpointer[t]; ++r){
               KKT_INT x=superrowindex[r]-supernodestart[s];
               assert(x>=0 && x<supersize);
               factorvalue[y+x] -= temp[tempindex];
               ++tempindex;
            }
         }
      }
      //   and finally solve with the diagonal block (which we factored above) to get the true factor column
      LAPACK::solve_symmetric_indefinite_with_factorization(supersize, supercolstart[s+1]-supercolstart[s],
                        diagvalue+diagblockstart[s], diagpiv+supernodestart[s], factorvalue+factorcolstart[s], info);
      // prepare to go on to the next supernode (updating column pointers to the next superrow)
      for(KKT_UINT p=0; p<factorrownonzeros.size(); ++p){
         KKT_INT t=factorrownonzeros[p]; // supernode index of this column within the row
         supercolpointer[t]=supercolnextpointer[t];
         DEBUG(if(supercolpointer[t]<supercolstart[t+1]) assert(superrowindex[supercolpointer[t]]>=supernodestart[s+1]));
      }
   }
   return pivots_are_all_fine;
}

// solve with supernodal factor ==============================================
void KKT_solve_with_supernodal_factorization(KKT_INT n,
                                             const KKT_INT *postordering,
                                             KKT_INT numsupernodes,
                                             const KKT_INT *supernodestart,
                                             const KKT_INT *supernodemap,
                                             const KKT_INT *supercolstart,
                                             const KKT_INT *superrowindex,
                                             const KKT_INT *factorcolstart,
                                             const double *factorvalue,
                                             const KKT_INT *diagblockstart,
                                             const double *diagvalue,
                                             const KKT_INT *diagpiv,
                                             KKT_INT nrhs,
                                             const double *rhs,
                                             double *solution)
{
   // first permute right-hand-side to chosen ordering, storing in y superblock by superblock
   // (also find size of largest supercolumn, maxnumrows)
   std::vector<double> y(n*nrhs);
   KKT_INT maxnumrows=0;
   for(KKT_INT s=0; s<numsupernodes; ++s){
      KKT_INT numrows=supercolstart[s+1]-supercolstart[s];
      if(numrows>maxnumrows) maxnumrows=numrows;
      KKT_INT blockstart=nrhs*supernodestart[s];
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      for(KKT_INT j=0; j<nrhs; ++j) for(KKT_INT i=0; i<supersize; ++i){
         assert(blockstart+supersize*j+i>=0 && blockstart+supersize*j+i<(KKT_INT)y.size());
         y[blockstart+supersize*j+i]=rhs[postordering[supernodestart[s]+i]+j*n];
      }
   }

   std::vector<double> temp(nrhs*maxnumrows); // used for intermediate matrix-matrix multiplies

   // then solve L*ynew=y in-place
   for(KKT_INT s=0; s<numsupernodes; ++s){
      KKT_INT numrows=supercolstart[s+1]-supercolstart[s];
      if(numrows==0) continue; // nothing to do
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      assert(numrows*supersize==factorcolstart[s+1]-factorcolstart[s]);
      // multiply the s'th block of the current solution by the subdiagonal part of this supercolumn of the factor
      //  the factor part is stored row-wise as a numrows*supersize dense matrix
      //  the right-hand-side part is stored columnwise as a supersize*nrhs dense matrix
      const double *y_s=&y[nrhs*supernodestart[s]];
      const double *factor_s=factorvalue+factorcolstart[s];
      assert(temp.size()>=(KKT_UINT)numrows*nrhs);
      BLAS::dgemm_TN(numrows, nrhs, supersize, factor_s, y_s, &temp[0]);
      // and scatter the result to update y
      //   this is somewhat involved due to block structure of y
      for(KKT_INT j=0; j<numrows; ++j){
         KKT_INT a=superrowindex[supercolstart[s]+j];
         KKT_INT t=supernodemap[a];
         KKT_INT tsize=supernodestart[t+1]-supernodestart[t];
         KKT_INT b=a-supernodestart[t];
         KKT_INT index=nrhs*supernodestart[t]+b;
         for(KKT_INT k=0; k<nrhs; ++k){
            assert(index+k*tsize>=0 && index+k*tsize<(KKT_INT)y.size());
            y[index+k*tsize]-=temp[j+k*numrows];
         }
      }
   }

   // then solve with D in-place, y:=inv(D)*y
   KKT_INT info;
   for(KKT_INT s=0; s<numsupernodes; ++s){
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      LAPACK::solve_symmetric_indefinite_with_factorization(supersize, nrhs,
                                                            diagvalue+diagblockstart[s], diagpiv+supernodestart[s],
                                                            &y[nrhs*supernodestart[s]], info);
   }

   // then solve L^T*ynew=y in-place
   for(KKT_INT s=numsupernodes-1; s>=0; --s){
      KKT_INT numrows=supercolstart[s+1]-supercolstart[s];
      if(numrows==0) continue; // nothing to do
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      // gather chunks of y into temp for dense matrix-matrix multiply
      //   so temp is a column-wise stored numrows*nrhs dense matrix)
      //   the factor part is stored row-wise as a numrows*supersize dense matrix
      //      (i.e. its transpose, which we want, is stored column-wise as a supersize*numrows matrix)
      for(KKT_INT j=0; j<numrows; ++j){
         KKT_INT a=superrowindex[supercolstart[s]+j];
         KKT_INT t=supernodemap[a];
         KKT_INT tsize=supernodestart[t+1]-supernodestart[t];
         KKT_INT b=a-supernodestart[t];
         KKT_INT index=nrhs*supernodestart[t]+b;
         for(KKT_INT k=0; k<nrhs; ++k){
            temp[j+k*numrows]=y[index+k*tsize];
         }
      }
      // and now update super block of y
      double *y_s=&y[nrhs*supernodestart[s]];
      const double *factor_s=factorvalue+factorcolstart[s];
      BLAS::dgemm_subNN(supersize, nrhs, numrows, factor_s, &temp[0], y_s);
   }

   // and repermute y to get solution
   for(KKT_INT s=0; s<numsupernodes; ++s){
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      KKT_INT blockstart=nrhs*supernodestart[s];
      for(KKT_INT j=0; j<nrhs; ++j) for(KKT_INT i=0; i<supersize; ++i){
         solution[j*n+postordering[supernodestart[s]+i]]=y[blockstart+j*supersize+i];
      }
   }
}

//=============================== signed Cholesky versions =======================================

int KKT_numerically_factor_supernodal_signed_cholesky(KKT_INT n,
                                                      const KKT_INT *colstart,
                                                      const KKT_INT *rowindex,
                                                      const double *value,
                                                      const KKT_INT *postordering,
                                                      const KKT_INT *inverseordering,
                                                      const KKT_INT *etreeparent,
                                                      const char *constrained,
                                                      // description of supernodes
                                                      KKT_INT numsupernodes,
                                                      const KKT_INT *supernodestart,
                                                      const KKT_INT *supernodemap,
                                                      // description of structure of factor (one supercolumn at a time)
                                                      const KKT_INT *supercolstart,
                                                      const KKT_INT *superrowindex,
                                                      // description of values in factor
                                                      const KKT_INT *factorcolstart,
                                                      double *factorvalue,
                                                      // description of values in diagonal
                                                      const KKT_INT *diagblockstart,
                                                      double *diagvalue)
{
   if(n<1) return 1;
   std::fill_n(factorvalue, factorcolstart[numsupernodes], (double)0); // zero out factor to begin
   int pivots_are_all_fine=1; // zero this flag if we encounter a dangerously small or zero pivot
   std::vector<KKT_INT> supercolpointer(numsupernodes); // keep track of where we are in structure of each column (progress row by row)
   std::copy(supercolstart, supercolstart+numsupernodes, supercolpointer.begin());
   std::vector<KKT_INT> supercolnextpointer(numsupernodes);
   std::vector<KKT_INT> factorrownonzeros; factorrownonzeros.reserve(numsupernodes);
   std::vector<KKT_INT> colstart_within_row; colstart_within_row.reserve(numsupernodes+1);
   std::vector<double> scaledrowvalue; scaledrowvalue.reserve(n);
   std::vector<double> diag_unfactored(diagblockstart[numsupernodes], 0); // where we store unfactored diagonal blocks
   std::vector<double> temp(sqr(n/numsupernodes)); // temporary storage used in various places...
   Marker marker(n), supermarker(numsupernodes);
   std::vector<KKT_INT> colmap(n); // for column of L under construction, maps nonzero rows to location in sparse structure
   for(KKT_INT s=0; s<numsupernodes; ++s){
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      // scatter the diagonal block of the original matrix into diag_unfactored
      for(KKT_INT a=supernodestart[s]; a<supernodestart[s+1]; ++a){
         KKT_INT i=postordering[a];
         for(KKT_INT p=colstart[i]; p<colstart[i+1]; ++p){
            KKT_INT j=rowindex[p];
            KKT_INT b=inverseordering[j];
            if(supernodemap[b]==s){
               KKT_INT location=diagblockstart[s] + (b-supernodestart[s])+supersize*(a-supernodestart[s]);
               diag_unfactored[location]=value[p];
            }
         }
      }
      // find supercolumn structure of this superrow using etree
      factorrownonzeros.resize(0);
      supermarker.clear();
      supermarker.set(s); // don't include diagonal block entries
      for(KKT_INT a=supernodestart[s]; a<supernodestart[s+1]; ++a){
         marker.clear();
         marker.set(a);
         KKT_INT i=postordering[a];
         for(KKT_INT p=colstart[i]; p<colstart[i+1]; ++p){
            KKT_INT j=rowindex[p];
            KKT_INT b=inverseordering[j];
            if(b>=a) continue;
            KKT_INT t=supernodemap[b];
            while(!marker[b]){
               marker.set(b);
               if(!supermarker[t]){
                  factorrownonzeros.push_back(t);
                  supermarker.set(t);
               }
               j=etreeparent[j];
               b=inverseordering[j];
               t=supernodemap[b];
            }
         }
      }
      // establish sparsity structure of this superrow (for each nonempty supercolumn, which rows are nonzero)
      colstart_within_row.resize(factorrownonzeros.size()+1);
      colstart_within_row[0]=0;
      KKT_INT nnz_in_superrow=0;
      for(KKT_UINT p=0; p<factorrownonzeros.size(); ++p){
         KKT_INT t=factorrownonzeros[p]; // supernode index of this column within the row
         assert(supercolpointer[t]==supercolstart[t+1] || superrowindex[supercolpointer[t]]>=supernodestart[s]);
         KKT_INT tsize=supernodestart[t+1]-supernodestart[t]; // how wide this supercolum is
         KKT_INT q;
         for(q=supercolpointer[t]; q<supercolstart[t+1] && superrowindex[q]<supernodestart[s+1]; ++q){ // look for which rows are nonempty
            assert(superrowindex[q]>=supernodestart[s]);
            nnz_in_superrow+=tsize;
         }
         supercolnextpointer[t]=q;
         colstart_within_row[p+1]=nnz_in_superrow;
      }
      // compute scaled superrow L(s,1:s-1)*D(1:s-1,1:s-1) in factorrownonzeros,
      // and subtract off inner product of superrow and scaled superrow from diag_unfactored
      scaledrowvalue.resize(nnz_in_superrow);
      for(KKT_UINT p=0; p<factorrownonzeros.size(); ++p){
         KKT_INT t=factorrownonzeros[p]; // supernode index of this column within the row
         KKT_INT tsize=supernodestart[t+1]-supernodestart[t]; // how wide supercolumn t is
         KKT_INT blocksize=colstart_within_row[p+1]-colstart_within_row[p]; // how many nonzeros in this chunk of the factor
         assert(blocksize%tsize==0);
         KKT_INT rsize=blocksize/tsize; // how many nonzero rows in this chunk
         // multiply block from factorvalue by diag_unfactored(t...) and store in scaledrowvalue
         //   factorvalue submatrix is at factorvalue+factorcolpointer_t, with dimensions rsize*tsize (stored row-wise)
         KKT_INT factorcolpointer_t=factorcolstart[t]+tsize*(supercolpointer[t]-supercolstart[t]);
         const double *factorsubmatrix=factorvalue+factorcolpointer_t;
         //   diag submatrix is at diag_unfactored+diagblockstart[t], with dimensions tsize*tsize (symmetric)
         const double *diagsubmatrix=&diag_unfactored[diagblockstart[t]];
         //   scaledrowvalue submatrix is at scaledrowvalue+colstart_within_row[p], with dimensions rsize*tsize (stored row-wise)
         double *scaledsubmatrix=&scaledrowvalue[colstart_within_row[p]];
         // Note: since factorsubmatrix and scaledsubmatrix are stored row-wise, not column-wise (as BLAS expect), this looks backwards --- but it isn't.
         BLAS::dgemm_NN(tsize, rsize, tsize, diagsubmatrix, factorsubmatrix, scaledsubmatrix);
         // multiply this part of the scaled row by the original row, storing in a temporary matrix
         temp.resize(rsize*rsize);
         BLAS::dgemm_TN(rsize, rsize, tsize, factorsubmatrix, scaledsubmatrix, &temp[0]);
         // and now subtract off that product from the D(s,s) unfactored diagonal block (a scattered/sparse update)
         for(KKT_INT q=0; q<rsize; ++q){
            KKT_INT x=superrowindex[supercolpointer[t]+q]-supernodestart[s];
            KKT_INT diagindex=diagblockstart[s]+x*supersize;
            for(KKT_INT r=0; r<rsize; ++r){
               KKT_INT y=superrowindex[supercolpointer[t]+r]-supernodestart[s];
               diag_unfactored[diagindex+y]-=temp[q*rsize+r];
            }
         }
      }
      // copy unfactored form to diagonal block, changing signs according to constrained (so result is positive definite)
      if(!constrained[postordering[supernodestart[s]]]){ // already positive
         std::copy(&diag_unfactored[diagblockstart[s]], &diag_unfactored[diagblockstart[s]]+sqr(supersize), diagvalue+diagblockstart[s]);
      }else{ // we have to flip the sign
         for(KKT_INT i=diagblockstart[s]; i<diagblockstart[s+1]; ++i){
            diagvalue[i]=-diag_unfactored[i];
         }
      }
      // factor diagonal block D(s,s)
      KKT_INT info;
      LAPACK::factor_symmetric_positive_definite(supersize, diagvalue+diagblockstart[s], info);
      if(info!=0){ // if there was a problem
         pivots_are_all_fine=0;
      }
      // figure out supercolumn s of factor, L(s+1:end,s) = (A(s+1:end,s)-L(s+1:end,1:s-1)*scaled_row^T)*inv(diag(s,s))
      //   first set up colmap (given a row, where the values are stored in the factor)
      for(KKT_INT p=supercolstart[s]; p<supercolstart[s+1]; ++p){
         colmap[superrowindex[p]]=factorcolstart[s]+(p-supercolstart[s])*supersize;
      }
      //   then copy entries from A into factor
      for(KKT_INT a=supernodestart[s]; a<supernodestart[s+1]; ++a){
         KKT_INT i=postordering[a];
         for(KKT_INT p=colstart[i]; p<colstart[i+1]; ++p){
            KKT_INT j=rowindex[p];
            KKT_INT b=inverseordering[j];
            if(supernodemap[b]>s){ // copy nonzeros below the diagonal superblock
               KKT_INT location=colmap[b]+(a-supernodestart[s]);
               assert(location>=factorcolstart[s] && location<factorcolstart[s+1]);
               factorvalue[location]=value[p];
            }
         }
      }
      //   and then subtract off factor*scaled_row^T update, one supercolumn at a time
      for(KKT_UINT p=0; p<factorrownonzeros.size(); ++p){
         KKT_INT t=factorrownonzeros[p]; // supernode index of this column within the row
         KKT_INT tsize=supernodestart[t+1]-supernodestart[t]; // how wide supercolumn t is
         KKT_INT blocksize=colstart_within_row[p+1]-colstart_within_row[p]; // how many nonzeros in supercolumn t of the (scaled) superrow
         KKT_INT rsize=blocksize/tsize; // how many nonzero rows in supercolumn t of the (scaled) superrow
         KKT_INT lsize=supercolstart[t+1]-supercolnextpointer[t]; // how many nonzero rows in the remainder of supercolumn t of the factor
         if(lsize==0) continue; // nothing to do here
         assert(supercolnextpointer[t]>=supercolpointer[t]);
         assert(supercolpointer[t]>=supercolstart[t]);
         assert(lsize<=n-supernodestart[s+1]);
         KKT_INT lstart=factorcolstart[t]+tsize*(supercolnextpointer[t]-supercolstart[t]); // where the nonzeros of the remainder of supercolumn in t begin
         assert(lstart<factorcolstart[t+1]);
         temp.resize(rsize*lsize);
         BLAS::dgemm_TN(rsize, lsize, tsize, &scaledrowvalue[colstart_within_row[p]], factorvalue+lstart, &temp[0]);
         // now scatter that product into the appropriate entries of supercolumn s
         KKT_INT tempindex=0;
         for(KKT_INT q=supercolnextpointer[t]; q<supercolstart[t+1]; ++q){
            KKT_INT y=colmap[superrowindex[q]];
            for(KKT_INT r=supercolpointer[t]; r<supercolnextpointer[t]; ++r){
               KKT_INT x=superrowindex[r]-supernodestart[s];
               assert(x>=0 && x<supersize);
               factorvalue[y+x] -= temp[tempindex];
               ++tempindex;
            }
         }
      }
      //   solve with the diagonal block (which we factored above) to get the true factor column
      LAPACK::solve_symmetric_positive_definite_with_factorization(supersize, supercolstart[s+1]-supercolstart[s],
                        diagvalue+diagblockstart[s], factorvalue+factorcolstart[s], info);
      //   and fix signs if necessary
      if(constrained[postordering[supernodestart[s]]]){
         for(KKT_INT i=factorcolstart[s]; i<factorcolstart[s+1]; ++i){
            factorvalue[i]=-factorvalue[i];
         }
      }
      // prepare to go on to the next supernode (updating column pointers to the next superrow)
      for(KKT_UINT p=0; p<factorrownonzeros.size(); ++p){
         KKT_INT t=factorrownonzeros[p]; // supernode index of this column within the row
         supercolpointer[t]=supercolnextpointer[t];
         DEBUG(if(supercolpointer[t]<supercolstart[t+1]) assert(superrowindex[supercolpointer[t]]>=supernodestart[s+1]));
      }
   }
   return pivots_are_all_fine;
}

void KKT_solve_with_supernodal_signed_cholesky(KKT_INT n,
                                               const char *constrained,
                                               const KKT_INT *postordering,
                                               KKT_INT numsupernodes,
                                               const KKT_INT *supernodestart,
                                               const KKT_INT *supernodemap,
                                               const KKT_INT *supercolstart,
                                               const KKT_INT *superrowindex,
                                               const KKT_INT *factorcolstart,
                                               const double *factorvalue,
                                               const KKT_INT *diagblockstart,
                                               const double *diagvalue,
                                               KKT_INT nrhs,
                                               const double *rhs,
                                               double *solution)
{
   // first permute right-hand-side to chosen ordering, storing in y superblock by superblock
   // (also find size of largest supercolumn, maxnumrows)
   std::vector<double> y(n*nrhs);
   KKT_INT maxnumrows=0;
   for(KKT_INT s=0; s<numsupernodes; ++s){
      KKT_INT numrows=supercolstart[s+1]-supercolstart[s];
      if(numrows>maxnumrows) maxnumrows=numrows;
      KKT_INT blockstart=nrhs*supernodestart[s];
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      for(KKT_INT j=0; j<nrhs; ++j) for(KKT_INT i=0; i<supersize; ++i){
         assert(blockstart+supersize*j+i>=0 && blockstart+supersize*j+i<(KKT_INT)y.size());
         y[blockstart+supersize*j+i]=rhs[postordering[supernodestart[s]+i]+j*n];
      }
   }

   std::vector<double> temp(nrhs*maxnumrows); // used for intermediate matrix-matrix multiplies

   // then solve L*ynew=y in-place
   for(KKT_INT s=0; s<numsupernodes; ++s){
      KKT_INT numrows=supercolstart[s+1]-supercolstart[s];
      if(numrows==0) continue; // nothing to do
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      assert(numrows*supersize==factorcolstart[s+1]-factorcolstart[s]);
      // multiply the s'th block of the current solution by the subdiagonal part of this supercolumn of the factor
      //  the factor part is stored row-wise as a numrows*supersize dense matrix
      //  the right-hand-side part is stored columnwise as a supersize*nrhs dense matrix
      const double *y_s=&y[nrhs*supernodestart[s]];
      const double *factor_s=factorvalue+factorcolstart[s];
      assert(temp.size()>=(KKT_UINT)numrows*nrhs);
      BLAS::dgemm_TN(numrows, nrhs, supersize, factor_s, y_s, &temp[0]);
      // and scatter the result to update y
      //   this is somewhat involved due to block structure of y
      for(KKT_INT j=0; j<numrows; ++j){
         KKT_INT a=superrowindex[supercolstart[s]+j];
         KKT_INT t=supernodemap[a];
         KKT_INT tsize=supernodestart[t+1]-supernodestart[t];
         KKT_INT b=a-supernodestart[t];
         KKT_INT index=nrhs*supernodestart[t]+b;
         for(KKT_INT k=0; k<nrhs; ++k){
            assert(index+k*tsize>=0 && index+k*tsize<(KKT_INT)y.size());
            y[index+k*tsize]-=temp[j+k*numrows];
         }
      }
   }

   // then solve with D in-place, y:=inv(D)*y
   KKT_INT info;
   for(KKT_INT s=0; s<numsupernodes; ++s){
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      LAPACK::solve_symmetric_positive_definite_with_factorization(supersize, nrhs, diagvalue+diagblockstart[s],
                                                                   &y[nrhs*supernodestart[s]], info);
   }

   // correct signs
   for(KKT_INT s=0; s<numsupernodes; ++s){
      if(constrained[postordering[supernodestart[s]]]){
         KKT_INT blockstart=nrhs*supernodestart[s];
         KKT_INT blockend=nrhs*supernodestart[s+1];
         for(KKT_INT j=blockstart; j<blockend; ++j){
            y[j]=-y[j];
         }
      }
   }

   // then solve L^T*ynew=y in-place
   for(KKT_INT s=numsupernodes-1; s>=0; --s){
      KKT_INT numrows=supercolstart[s+1]-supercolstart[s];
      if(numrows==0) continue; // nothing to do
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      // gather chunks of y into temp for dense matrix-matrix multiply
      //   so temp is a column-wise stored numrows*nrhs dense matrix)
      //   the factor part is stored row-wise as a numrows*supersize dense matrix
      //      (i.e. its transpose, which we want, is stored column-wise as a supersize*numrows matrix)
      for(KKT_INT j=0; j<numrows; ++j){
         KKT_INT a=superrowindex[supercolstart[s]+j];
         KKT_INT t=supernodemap[a];
         KKT_INT tsize=supernodestart[t+1]-supernodestart[t];
         KKT_INT b=a-supernodestart[t];
         KKT_INT index=nrhs*supernodestart[t]+b;
         for(KKT_INT k=0; k<nrhs; ++k){
            temp[j+k*numrows]=y[index+k*tsize];
         }
      }
      // and now update super block of y
      double *y_s=&y[nrhs*supernodestart[s]];
      const double *factor_s=factorvalue+factorcolstart[s];
      BLAS::dgemm_subNN(supersize, nrhs, numrows, factor_s, &temp[0], y_s);
   }

   // and repermute y to get solution
   for(KKT_INT s=0; s<numsupernodes; ++s){
      KKT_INT supersize=supernodestart[s+1]-supernodestart[s];
      KKT_INT blockstart=nrhs*supernodestart[s];
      for(KKT_INT j=0; j<nrhs; ++j) for(KKT_INT i=0; i<supersize; ++i){
         solution[j*n+postordering[supernodestart[s]+i]]=y[blockstart+j*supersize+i];
      }
   }
}

