#ifndef KKTSUPPORT_H
#define KKTSUPPORT_H

#include <cassert>
#include <cmath>
#include <cstring>
#include <algorithm>
#include <iostream>
#include <vector>
#include "kktdirect.h"

//================================ debugging =================================

#ifdef NDEBUG
#define DEBUG(x)
#else
#define DEBUG(x) x
#endif

//================================= frequent operations ======================

inline void remove_from_list(std::vector<KKT_INT> &list, KKT_INT &index)
{
   list[index]=list.back();
   list.pop_back();
   --index;
}

template<class T>
T sqr(T x)
{ return x*x; }

//================================= Hash Table ===============================
// Very very simple - no deletion even

struct HashTable
{
   HashTable(KKT_INT table_size)
      : table(table_size)
   {}

   void add(KKT_INT key, KKT_INT value)
   {
      KKT_INT h=key%table.size();
      if(h<0) h+=table.size();
      assert(h>=0 && h<(KKT_INT)table.size());
      table[h].push_back(std::pair<KKT_INT,KKT_INT>(key,value));
   }

   void find(KKT_INT key, std::vector<KKT_INT> &values) const
   {
      values.resize(0);
      KKT_INT h=key%table.size();
      for(KKT_UINT a=0; a<table[h].size(); ++a){
         if(table[h][a].first==key)
            values.push_back(table[h][a].second);
      }
   }

   protected:
   std::vector<std::vector<std::pair<KKT_INT,KKT_INT> > > table;
};

//================================= Trie =====================================
// Stores a KKT_INT value for each non-negative 32-bit key, at most one value per key
// (if KKT_UINTs are larger than 32-bit, only the lowest 32 bits of keys will be used,
//  and everything above 2^32-1 is treated as equal to 2^32-1).
// Optimized for find, add, and remove; find_min is still O(1) but not as fast as it could be.

struct Trie32
{
   Trie32(std::vector<KKT_UINT> &key_)
      : key(key_)
   { std::fill_n(trie, R, (KKT_INT***)0); }

   ~Trie32(void)
   {
      for(KKT_UINT a=0; a<R; ++a){
         if(trie[a]){
            for(KKT_UINT b=0; b<R; ++b){
               if(trie[a][b]){
                  for(KKT_UINT c=0; c<R; ++c){
                     if(trie[a][b][c]){
                        delete[] trie[a][b][c];
                        trie[a][b][c]=0;
                     }
                  }
                  delete[] trie[a][b];
                  trie[a][b]=0;
               }
            }
            delete[] trie[a];
            trie[a]=0;
         }
      }
   }

   KKT_INT find_min(void) const
   {
      KKT_UINT a;
      for(a=0; a<R && !trie[a]; ++a)
         ;
      if(a==R) return -1;
      KKT_UINT b;
      for(b=0; b<R && !trie[a][b]; ++b)
         ;
      assert(b<R);
      KKT_UINT c;
      for(c=0; c<R && !trie[a][b][c]; ++c)
         ;
      assert(c<R);
      KKT_UINT d;
      for(d=0; d<R && trie[a][b][c][d]==-1; ++d)
         ;
      assert(d<R);
      DEBUG(std::cerr<<"Find min: "<<a<<" "<<b<<" "<<c<<" "<<d<<": "<<trie[a][b][c][d]<<std::endl);
      return trie[a][b][c][d];
   }

   KKT_INT find(KKT_UINT k) const
   {
#ifdef USE_LONG_LONG_INT
      if(k>0xffffffffUL) k=0xffffffffUL;
#endif
      KKT_UINT a=(k>>(3*Rshift)) & Rbits;
      if(!trie[a]) return -1;
      KKT_UINT b=(k>>(2*Rshift)) & Rbits;
      if(!trie[a][b]) return -1;
      KKT_UINT c=(k>>Rshift) & Rbits;
      if(!trie[a][b][c]) return -1;
      KKT_UINT d=k & Rbits;
      return trie[a][b][c][d];
   }

   void add(KKT_INT i)
   {
      KKT_UINT k=key[i];
#ifdef USE_LONG_LONG_INT
      if(k>0xffffffffUL) k=0xffffffffUL;
#endif
      KKT_UINT a=(k>>(3*Rshift)) & Rbits;
      if(!trie[a]){
         trie[a]=new KKT_INT**[R];
         std::fill_n(trie[a], R, (KKT_INT**)0);
      }
      KKT_UINT b=(k>>(2*Rshift)) & Rbits;
      if(!trie[a][b]){
         trie[a][b]=new KKT_INT*[R];
         std::fill_n(trie[a][b], R, (KKT_INT*)0);
      }
      KKT_UINT c=(k>>Rshift) & Rbits;
      if(!trie[a][b][c]){
         trie[a][b][c]=new KKT_INT[R+1]; // extra entry for count of how many in here
         std::fill_n(trie[a][b][c], R, -1);
         trie[a][b][c][R]=0;
      }
      KKT_UINT d=k & Rbits;
      assert(trie[a][b][c][d]==-1); // we don't handle multiple entries with the same key being added!
      trie[a][b][c][d]=i;
      ++trie[a][b][c][R];
   }

   void remove(KKT_INT i)
   {
      KKT_UINT k=key[i];
#ifdef USE_LONG_LONG_INT
      if(k>0xffffffffUL) k=0xffffffffUL;
#endif
      KKT_UINT a=(k>>(3*Rshift)) & Rbits;
      KKT_UINT b=(k>>(2*Rshift)) & Rbits;
      KKT_UINT c=(k>>Rshift) & Rbits;
      KKT_UINT d=k & Rbits;
      assert(trie[a] && trie[a][b] && trie[a][b][c] && trie[a][b][c][d]!=-1);
      trie[a][b][c][d]=-1;
      assert(trie[a][b][c][R]>0);
      --trie[a][b][c][R]; // decrement count of entries in trie[a][b][c]
      if(trie[a][b][c][R]==0){ // if there are none left, clean up
         delete[] trie[a][b][c];
         trie[a][b][c]=0;

         bool empty_next_level=true;
         for(KKT_UINT c=0; c<R; ++c){
            if(trie[a][b][c]){
               empty_next_level=false;
               break;
            }
         }
         if(!empty_next_level) return;
         delete[] trie[a][b];
         trie[a][b]=0;

         empty_next_level=true;
         for(KKT_UINT b=0; b<R; ++b){
            if(trie[a][b]){
               empty_next_level=false;
               break;
            }
         }
         if(!empty_next_level) return;
         delete[] trie[a];
         trie[a]=0;
      }
   }

   protected:
   std::vector<KKT_UINT> &key;
   static const KKT_UINT R=256;
   static const KKT_UINT Rshift=8;
   static const KKT_UINT Rbits=R-1;
   KKT_INT ***trie[R];
};

//============================= Priority Queue ===============================
// Note: if KKT_UNIT is 64-bit, then only the lower 32-bits are used; every key
// above 2^32-1 will be treated as equal to 2^32-1 --- thus if there are no
// keys below 2^32-1, find_min() may not actually return an index with minimum key.

struct PriorityQueue
{
   PriorityQueue(std::vector<KKT_UINT> &key_)
      :  key(key_),
         member(key.size(), false),
         count(0),
         keyhead(key),
         next(key.size(),-1), prev(key.size(),-1)
   {}

   bool empty(void) const
   { return count==0; }

   KKT_INT find_min(void) const
   {
      assert(!empty());
      return keyhead.find_min();
   }

   void add(KKT_INT i)
   {
      assert(i>=0 && i<(KKT_INT)member.size());
      assert(!member[i]);
      KKT_INT h=keyhead.find(key[i]);
      if(h==-1){
         keyhead.add(i);
      }else{
         assert(h>=0 && h<(KKT_INT)next.size());
         next[i]=next[h];
         if(next[i]!=-1) prev[next[i]]=i;
         prev[i]=h;
         next[h]=i;
      }
      member[i]=true;
      ++count;
   }

   void modify_key(KKT_INT i, KKT_UINT newkey)
   {
      assert(i>=0 && i<(KKT_INT)member.size());
      if(newkey==key[i]) return;
      remove(i);
      key[i]=newkey;
      add(i);
   }

   void remove(KKT_INT i)
   {
      assert(i>=0 && i<(KKT_INT)member.size());
      assert(member[i]);
      if(prev[i]!=-1){
         next[prev[i]]=next[i];
         if(next[i]!=-1){
            prev[next[i]]=prev[i];
         }
      }else{ // it's the head of a list
         keyhead.remove(i);
         if(next[i]!=-1){
            keyhead.add(next[i]);
            prev[next[i]]=-1;
         }
      }
      member[i]=false;
      next[i]=-1;
      prev[i]=-1;
      --count;
   }

   bool contains(KKT_INT i) const
   { return member[i]; }

   protected:
   std::vector<KKT_UINT> &key;
   std::vector<bool> member; // for each node, whether or not it's in the priority queue
   KKT_INT count; // how many nodes contained in the pq
   Trie32 keyhead; // for each stored key, the start of the list for that key
   std::vector<KKT_INT> next, prev; // for doubly linked list buckets, one for each key value
                                // (keyhead points to the first in the list). -1 used to mark beginning/end.
};

//===================================== Marker ===============================
// Essentially a bit vector with an efficient means for clearing all bits

struct Marker
{
   Marker(KKT_INT n)
      : flag(n,0), counter(1)
   {}

   void clear(void)
   {
      ++counter;
      if(counter==0){ // if we hit wrap-around, reset everything
         for(KKT_UINT i=0; i<flag.size(); ++i)
            flag[i]=0;
         counter=1;
      }
   }

   bool operator[](KKT_INT i) const
   { return flag[i]==counter; }

   void set(KKT_INT i)
   { flag[i]=counter; }

   void swap(Marker &othermarker)
   {
      std::swap(flag, othermarker.flag);
      std::swap(counter, othermarker.counter);
   }

   protected:
   std::vector<KKT_UINT> flag;
   KKT_UINT counter;
};

#endif
