#include <algorithm>
#include "kktdirect.h"
#include "kktsupport.h"

void KKT_detect_constraints(KKT_INT n, 
                            const KKT_INT *colstart,
                            const KKT_INT *rowindex,
                            const double *value,
                            char *constrained)
{
   for(KKT_INT i=0; i<n; ++i){
      constrained[i]=1;
      for(KKT_INT a=colstart[i]; a<colstart[i+1]; ++a){
         if(rowindex[a]==i && (!value || value[a]>0)){
            constrained[i]=0;
            break;
         }
      }
   }
}

// calculates elimination tree given ordering and its inverse permutation.
// parent[] is the standard description (for each node, what is its parent in the tree, -1 if it is a root)
// children[i] is a list of the children of node i in the etree, *IN THE ORIGINAL ORDER OF ELIMINATION*.
static void KKT_etree(KKT_INT n, 
                      const KKT_INT *colstart,
                      const KKT_INT *rowindex,
                      const KKT_INT *ordering,
                      const KKT_INT *inv, // inverse of ordering
                      KKT_INT *parent,
                      std::vector<std::vector<KKT_INT> > &children)
{
   if(n<1) return;
   for(KKT_INT i=0; i<n; ++i) parent[i]=-1;
   std::vector<KKT_INT> ancestor(n,-1); // short-cut version of parent
   children.resize(0); children.resize(n);
   for(KKT_INT a=0; a<n; ++a){
      KKT_INT i=ordering[a]; // i is the a'th node in ordering
      assert(inv[i]==a);
      for(KKT_INT b=colstart[i]; b<colstart[i+1]; ++b){
         KKT_INT j=rowindex[b]; // j is a neighbour of i in matrix
         KKT_INT c=inv[j];  // i.e. j was the c'th node in ordering
         if(c<a){ // if j was ordered before i
            KKT_INT k=j;
            while(ancestor[k]!=-1){
               KKT_INT next_k=ancestor[k];
               if(next_k==i) goto skip_this_neighbour;
               assert(next_k!=k);
               ancestor[k]=i;
               k=next_k;
               assert(inv[k]<a);
            }
            assert(parent[k]==-1);
            parent[k]=i;
            ancestor[k]=i;
            children[i].push_back(inv[k]); // NOTE: we put k's rank in the ordering, not k itself (will fix this later)
         }
         skip_this_neighbour: ;
      }
   }
   for(KKT_INT i=0; i<n; ++i){
      sort(children[i].begin(), children[i].end()); // sort children by occurrence in ordering
      for(KKT_UINT a=0; a<children[i].size(); ++a){
         children[i][a]=ordering[children[i][a]]; // translate back to actual node labels
         assert(parent[children[i][a]]==i);
      }
   }
}

void KKT_postorder(KKT_INT n, 
                   const KKT_INT *colstart,
                   const KKT_INT *rowindex,
                   const KKT_INT *ordering,
                   KKT_INT *postordering,
                   KKT_INT *inverseordering,
                   KKT_INT *etreeparent)
{
   if(n<1) return;
   // inverse of initial ordering
   for(KKT_INT a=0; a<n; ++a){
      inverseordering[ordering[a]]=a;
   }
   // get the initial ordering's elimination tree (actually, a forest)
   std::vector<std::vector<KKT_INT> > children(n);
   KKT_etree(n, colstart, rowindex, ordering, inverseordering, etreeparent, children);
   // depth-first-search from each root (creating postordering in reverse)
   std::vector<KKT_INT> stack;
   stack.reserve(n);
   KKT_INT k=n; // where we write into postordering
   DEBUG(std::vector<bool> postordered(n,false));
   for(KKT_INT r=n-1; r>=0; --r){
      if(etreeparent[r]!=-1) continue; // just look for the roots of the elimination "tree" (really, forest)
      stack.push_back(r);
      while(!stack.empty()){
         KKT_INT i=stack.back(); stack.pop_back();
         assert(!postordered[i]);
         assert(k>0);
         --k; postordering[k]=i;
         DEBUG(postordered[i]=true);
         for(KKT_INT b=children[i].size()-1; b>=0; --b){
            stack.push_back(children[i][b]);
         }
      }
   }
   // now update inverse_ordering to match postordering (note that etree_parent is unchanged)
   for(KKT_INT a=0; a<n; ++a){
      inverseordering[postordering[a]]=a;
   }
}

int KKT_check_constraints_satisified_by_ordering(KKT_INT n,
                                                 const KKT_INT *colstart,
                                                 const KKT_INT *rowindex,
                                                 const char *constrained,
                                                 const KKT_INT *ordering)
{
   if(n<1) return 1;
   std::vector<bool> ordered(n,false);
   for(KKT_INT a=0; a<n; ++a){
      KKT_INT i=ordering[a];
      ordered[i]=true;
      if(constrained[i]){
         for(KKT_INT b=colstart[i]; b<colstart[i+1]; ++b){
            KKT_INT j=rowindex[b];
            if(!constrained[j] && !ordered[j]) return 0;
         }
      }
   }
   return 1;
}

// efficiently computes the number of nonzeros in each column of L (in L's postorder,
// not the original ordering of the matrix) including the diagonal
static void KKT_compute_column_count(KKT_INT n,
                                     const KKT_INT *colstart,
                                     const KKT_INT *rowindex,
                                     const KKT_INT *postordering,
                                     const KKT_INT *inverseordering,
                                     const KKT_INT *etreeparent,
                                     KKT_INT *columncount)
{
   if(n<1) return;
   // initially, we'll use columncount to store the "delta" components for efficiently accumulating column counts
   // set up "first" --- array of first descendent (according to postordering) of each node in elimination tree
   // and for those nodes that are leaves, initialize delta to 1 instead of 0
   std::fill_n(columncount, n, 0);
   std::vector<KKT_INT> first(n, -1);
   for(KKT_INT a=0; a<n; ++a){
      if(first[a]==-1) columncount[a]=1;
      for(KKT_INT b=a; first[b]==-1; ){
         first[b]=a;
         if(etreeparent[postordering[b]]==-1) break;
         else b=inverseordering[etreeparent[postordering[b]]];
      }
   }
   // now figure out delta
   std::vector<KKT_INT> ancestor(n), maxfirst(n, -1), prevleaf(n, -1);
   for(KKT_INT a=0; a<n; ++a) ancestor[a]=a;
   for(KKT_INT a=0; a<n; ++a){
      KKT_INT i=postordering[a];
      if(etreeparent[i]!=-1)
         --columncount[inverseordering[etreeparent[i]]];
      for(KKT_INT b=colstart[i]; b<colstart[i+1]; ++b){
         KKT_INT j=rowindex[b];
         KKT_INT c=inverseordering[j];
         if(c<=a || first[a]<=maxfirst[c]) continue;
         maxfirst[c]=first[a];
         KKT_INT jprev=prevleaf[c];
         prevleaf[c]=a;
         if(jprev==-1){
            ++columncount[a];
         }else{
            KKT_INT d=jprev;
            while(d!=ancestor[d]) d=ancestor[d];
            KKT_INT e=jprev;
            while(e!=d){
               KKT_INT nexte=ancestor[e];
               ancestor[e]=d;
               e=nexte;
            }
            ++columncount[a];
            --columncount[d];
         }
      }
      if(etreeparent[i]!=-1)
         ancestor[a]=inverseordering[etreeparent[i]];
   }
   // and accumulate into the true column counts
   for(KKT_INT a=0; a<n; ++a){
      KKT_INT p=etreeparent[postordering[a]];
      if(p!=-1) columncount[inverseordering[p]]+=columncount[a];
   }
}

void KKT_symbolically_factor_uplooking(KKT_INT n,
                                       const KKT_INT *colstart,
                                       const KKT_INT *rowindex,
                                       const KKT_INT *postordering,
                                       const KKT_INT *inverseordering,
                                       const KKT_INT *etreeparent,
                                       KKT_INT *factorcolstart)
{
   if(n<1) return;
   KKT_compute_column_count(n, colstart, rowindex, postordering, inverseordering, etreeparent, factorcolstart+1);
   // do the cumulative sum to get where each column of the factor actually starts
   factorcolstart[0]=0;
   for(KKT_INT a=1; a<n+1; ++a){
      factorcolstart[a]+=factorcolstart[a-1]-1; // -1 since we don't include the unit diagonal in this case
   }
}

void KKT_symbolically_factor_signed_cholesky(KKT_INT n,
                                             const KKT_INT *colstart,
                                             const KKT_INT *rowindex,
                                             const KKT_INT *postordering,
                                             const KKT_INT *inverseordering,
                                             const KKT_INT *etreeparent,
                                             KKT_INT *factorcolstart)
{
   if(n<1) return;
   KKT_compute_column_count(n, colstart, rowindex, postordering, inverseordering, etreeparent, factorcolstart+1);
   // do the cumulative sum to get where each column of the factor actually starts
   for(KKT_INT a=1; a<n+1; ++a){
      factorcolstart[a]+=factorcolstart[a-1];
   }
}

void KKT_symbolically_factor_supernodal_initial(KKT_INT n,
                                                const KKT_INT *colstart,
                                                const KKT_INT *rowindex,
                                                const KKT_INT *postordering,
                                                const KKT_INT *inverseordering,
                                                const KKT_INT *etreeparent,
                                                KKT_INT max_supernode_size,
                                                KKT_INT *numsupernodes,
                                                KKT_INT *supernodestart,
                                                KKT_INT *supernodemap,
                                                KKT_INT *supercolstart,
                                                KKT_INT *factorcolstart,
                                                KKT_INT *diagblockstart)
{
   if(n<1) return;
   // first get column counts (using diagblockstart as a temporary)
   KKT_INT *columncount=diagblockstart+1; // NOTE: alias for readability!
   KKT_compute_column_count(n, colstart, rowindex, postordering, inverseordering, etreeparent, columncount);
   // then determine supernodes
   *numsupernodes=0;
   supernodestart[0]=0;
   while(supernodestart[*numsupernodes]<n){
      KKT_INT a=supernodestart[*numsupernodes];
      do{
         ++a;
      }while(a<n && etreeparent[postordering[a-1]]==postordering[a] && columncount[a-1]==columncount[a]+1);
      KKT_INT supernode_size=a-supernodestart[*numsupernodes];
      assert(supernode_size>0);
      if(supernode_size<=max_supernode_size){
         ++*numsupernodes;
         supernodestart[*numsupernodes]=a;
      }else{ // otherwise lets split it up roughly evenly
         KKT_INT m=(KKT_INT)std::ceil(supernode_size/(double)max_supernode_size); // use m sub-supernodes
         KKT_INT start=supernodestart[*numsupernodes];
         for(int i=0; i<m; ++i){
            ++*numsupernodes;
            supernodestart[*numsupernodes]=start+(KKT_INT)(((i+1)*(double)supernode_size)/m);
            assert(supernodestart[*numsupernodes]>supernodestart[*numsupernodes-1]);
         }
         assert(supernodestart[*numsupernodes]==start+supernode_size);
      }
      assert(supernodestart[*numsupernodes]==a);
   }
   KKT_INT node=0;
   for(KKT_INT a=0; a<n; ++a){
      if(a>=supernodestart[node+1]) ++node;
      supernodemap[a]=node;
   }
   // figure out layout of factor in memory (column starts)
   // and the diagonal block sizes, overwriting the aliased columncount as we go
   supercolstart[0]=0;
   factorcolstart[0]=0;
   diagblockstart[0]=0;
   for(KKT_INT a=0; a<*numsupernodes; ++a){
      KKT_INT supernode_size=supernodestart[a+1]-supernodestart[a];
      KKT_INT supercolumn_size=columncount[supernodestart[a]]-supernode_size;
      supercolstart[a+1]=supercolstart[a]+supercolumn_size;
      factorcolstart[a+1]=factorcolstart[a]+supernode_size*supercolumn_size;
      diagblockstart[a+1]=diagblockstart[a]+supernode_size*supernode_size;
   }
}

void KKT_symbolically_factor_supernodal_initial_for_signed_cholesky(KKT_INT n,
                                                                    const KKT_INT *colstart,
                                                                    const KKT_INT *rowindex,
                                                                    const char *constrained,
                                                                    KKT_INT max_supernode_size,
                                                                    KKT_INT *postordering,
                                                                    KKT_INT *inverseordering,
                                                                    KKT_INT *etreeparent,
                                                                    KKT_INT *numsupernodes,
                                                                    KKT_INT *supernodestart,
                                                                    KKT_INT *supernodemap,
                                                                    KKT_INT *supercolstart,
                                                                    KKT_INT *factorcolstart,
                                                                    KKT_INT *diagblockstart)
{
   if(n<1) return;
   // first get column counts (using diagblockstart as a temporary)
   KKT_INT *columncount=diagblockstart+1; // NOTE: alias for readability!
   KKT_compute_column_count(n, colstart, rowindex, postordering, inverseordering, etreeparent, columncount);
   // then determine supernodes
   *numsupernodes=0;
   supernodestart[0]=0;
   std::vector<KKT_INT> local_unconstrained, local_constrained;
   while(supernodestart[*numsupernodes]<n){
      KKT_INT a=supernodestart[*numsupernodes];
      local_unconstrained.resize(0);
      local_constrained.resize(0);
      if(constrained[postordering[a]]){ // if it starts with a constrained node, continue this supernode only with constrained nodes
         do{
            local_constrained.push_back(postordering[a]);
            ++a;
         }while(a<n && constrained[postordering[a]] && etreeparent[postordering[a-1]]==postordering[a] && columncount[a-1]==columncount[a]+1);
      }else{ // starts with unconstrained, so we're free to reorder rest of raw supernode
         do{
            if(constrained[postordering[a]]) local_constrained.push_back(postordering[a]);
            else local_unconstrained.push_back(postordering[a]);
            ++a;
         }while(a<n && etreeparent[postordering[a-1]]==postordering[a] && columncount[a-1]==columncount[a]+1);
         // modify postordering to order all the unconstrained ones first
         if(!local_constrained.empty()){
            KKT_INT b=supernodestart[*numsupernodes];
            for(KKT_INT p=0; p<(KKT_INT)local_unconstrained.size(); ++p, ++b) postordering[b]=local_unconstrained[p];
            for(KKT_INT p=0; p<(KKT_INT)local_constrained.size(); ++p, ++b)   postordering[b]=local_constrained[p];
            assert(b==a);
         }
      }
      // now set up the supernodes
      if(!local_unconstrained.empty()){
         if((KKT_INT)local_unconstrained.size()<=max_supernode_size){
            ++*numsupernodes;
            supernodestart[*numsupernodes]=supernodestart[*numsupernodes-1]+local_unconstrained.size();
         }else{ // otherwise lets split it up roughly evenly
            KKT_INT m=(KKT_INT)std::ceil(local_unconstrained.size()/(double)max_supernode_size); // use m sub-supernodes
            KKT_INT start=supernodestart[*numsupernodes];
            for(int i=0; i<m; ++i){
               ++*numsupernodes;
               supernodestart[*numsupernodes]=start+(KKT_INT)(((i+1)*(double)local_unconstrained.size())/m);
               assert(supernodestart[*numsupernodes]>supernodestart[*numsupernodes-1]);
            }
         }
      }
      if(!local_constrained.empty()){
         if((KKT_INT)local_constrained.size()<=max_supernode_size){
            ++*numsupernodes;
            supernodestart[*numsupernodes]=supernodestart[*numsupernodes-1]+local_constrained.size();
         }else{ // otherwise lets split it up roughly evenly
            KKT_INT m=(KKT_INT)std::ceil(local_constrained.size()/(double)max_supernode_size); // use m sub-supernodes
            KKT_INT start=supernodestart[*numsupernodes];
            for(int i=0; i<m; ++i){
               ++*numsupernodes;
               supernodestart[*numsupernodes]=start+(KKT_INT)(((i+1)*(double)local_constrained.size())/m);
               assert(supernodestart[*numsupernodes]>supernodestart[*numsupernodes-1]);
            }
         }
      }
      assert(supernodestart[*numsupernodes]==a);
   }
   // get the associated map
   KKT_INT node=0;
   for(KKT_INT a=0; a<n; ++a){
      if(a>=supernodestart[node+1]) ++node;
      supernodemap[a]=node;
   }
   // figure out layout of factor in memory (column starts)
   // and the diagonal block sizes, overwriting the aliased columncount as we go
   supercolstart[0]=0;
   factorcolstart[0]=0;
   diagblockstart[0]=0;
   for(KKT_INT a=0; a<*numsupernodes; ++a){
      KKT_INT supernode_size=supernodestart[a+1]-supernodestart[a];
      KKT_INT supercolumn_size=columncount[supernodestart[a]]-supernode_size;
      supercolstart[a+1]=supercolstart[a]+supercolumn_size;
      factorcolstart[a+1]=factorcolstart[a]+supernode_size*supercolumn_size;
      diagblockstart[a+1]=diagblockstart[a]+supernode_size*supernode_size;
   }
   // last of all, fix up inverseordering and elimination tree (which may have been modified by reordering)
   // There's undoubtedly a smart way to update, but for now, just recompute from scratch
   for(KKT_INT a=0; a<n; ++a) inverseordering[postordering[a]]=a;
   std::vector<std::vector<KKT_INT> > children(n);
   KKT_etree(n, colstart, rowindex, postordering, inverseordering, etreeparent, children);
}

// form a compressed version of the elimination tree, where the vertices are supernodes
static void KKT_compress_elimination_tree_with_supernodes(KKT_INT n,
                                                          const KKT_INT *postordering,
                                                          const KKT_INT *inverseordering,
                                                          const KKT_INT *etreeparent,
                                                          KKT_INT numsupernodes,
                                                          const KKT_INT *supernodestart,
                                                          const KKT_INT *supernodemap,
                                                          KKT_INT *superparent)
{
   for(KKT_INT a=0; a<numsupernodes; ++a){
      KKT_INT i=postordering[supernodestart[a+1]-1]; // real vertex label of last node in supernode a
      KKT_INT p=etreeparent[i];
      if(p==-1) // is this a root?
         superparent[a]=-1;
      else
         superparent[a]=supernodemap[inverseordering[p]];
   }
}

void KKT_symbolically_factor_supernodal_final(KKT_INT n,
                                              const KKT_INT *colstart,
                                              const KKT_INT *rowindex,
                                              const KKT_INT *postordering,
                                              const KKT_INT *inverseordering,
                                              const KKT_INT *etreeparent,
                                              KKT_INT numsupernodes,
                                              const KKT_INT *supernodestart,
                                              const KKT_INT *supernodemap,
                                              const KKT_INT *supercolstart,
                                              KKT_INT *superrowindex)
{
   if(n<1) return;
   std::vector<KKT_INT> superparent(numsupernodes,-1);
   KKT_compress_elimination_tree_with_supernodes(n, postordering, inverseordering, etreeparent,
                                                 numsupernodes, supernodestart, supernodemap,
                                                 &superparent[0]);
   // fill in structure row by row (use superparent to quickly determine which supercolumns are nonzero in each row)
   std::vector<KKT_INT> supercolend(numsupernodes);
   std::copy(supercolstart, supercolstart+numsupernodes, supercolend.begin());
   Marker supermarker(numsupernodes);
   for(KKT_INT a=0; a<n; ++a){
      supermarker.clear();
      supermarker.set(supernodemap[a]); // don't include entries in the diagonal block
      KKT_INT i=postordering[a];
      for(KKT_INT p=colstart[i]; p<colstart[i+1]; ++p){
         KKT_INT j=rowindex[p];
         KKT_INT b=inverseordering[j];
         if(b>=a) continue;
         KKT_INT s=supernodemap[b];
         while(!supermarker[s]){
            KKT_INT q=supercolend[s];
            ++supercolend[s];
            superrowindex[q]=a;
            supermarker.set(s);
            s=superparent[s];
            assert(s>=0);
         }
      }
   }
   DEBUG(
   for(KKT_INT s=0; s<numsupernodes; ++s){
      assert(supercolend[s]==supercolstart[s+1]);
   })
}

